from exegol.console.cli.ExegolCompleter import HybridContainerImageCompleter, VoidCompleter, BuildProfileCompleter
from exegol.console.cli.actions.Command import Command, Option, GroupArg
from exegol.console.cli.actions.GenericParameters import ContainerCreation, ContainerSpawnShell, ContainerMultiSelector, ContainerSelector, ImageSelector, ImageMultiSelector, ContainerStart
from exegol.manager.ExegolManager import ExegolManager
from exegol.utils.ExeLog import logger


class Start(Command, ContainerCreation, ContainerSpawnShell):
    """Automatically create, start / resume and enter an Exegol container"""

    def __init__(self) -> None:
        Command.__init__(self)
        ContainerCreation.__init__(self, self.groupArgs)
        ContainerSpawnShell.__init__(self, self.groupArgs)

        self._usages = {
            "Start interactively a container": "exegol start",
            "Create a [blue]demo[/blue] container using [bright_blue]full[/bright_blue] image": "exegol start [blue]demo[/blue] [bright_blue]full[/bright_blue]",
            "Spawn a shell from [blue]demo[/blue] container": "exegol start [blue]demo[/blue]",
            "Create a container [blue]test[/blue] with a custom shared workspace": "exegol start [blue]test[/blue] [bright_blue]full[/bright_blue] -w [magenta]./project/pentest/[/magenta]",
            "Create a container [blue]test[/blue] sharing the current working directory": "exegol start [blue]test[/blue] [bright_blue]full[/bright_blue] -cwd",
            "Create a container [blue]htb[/blue] with a VPN": "exegol start [blue]htb[/blue] [bright_blue]full[/bright_blue] --vpn [magenta]~/vpn/[/magenta][bright_magenta]lab_Dramelac.ovpn[/bright_magenta]",
            "Create a container [blue]app[/blue] with custom volume": "exegol start [blue]app[/blue] [bright_blue]full[/bright_blue] -V [bright_magenta]/var/app/[/bright_magenta]:[bright_magenta]/app/[/bright_magenta]",
            "Create a container [blue]app[/blue] with custom volume in [blue]ReadOnly[/blue]": "exegol start [blue]app[/blue] [bright_blue]full[/bright_blue] -V [bright_magenta]/var/app/[/bright_magenta]:[bright_magenta]/app/[/bright_magenta]:[blue]ro[/blue]",
            "Get a [blue]tmux[/blue] shell": "exegol start --shell [blue]tmux[/blue]",
            "Share a specific [blue]hardware device[/blue] [bright_black](e.g. Proxmark)[/bright_black]": "exegol start -d /dev/ttyACM0",
            "Share every [blue]USB device[/blue] connected to the host": "exegol start -d /dev/bus/usb/",
        }

    def __call__(self, *args, **kwargs):
        return ExegolManager.start


class Stop(Command, ContainerMultiSelector):
    """Stop an Exegol container"""

    def __init__(self) -> None:
        Command.__init__(self)
        ContainerMultiSelector.__init__(self, self.groupArgs)

        self._usages = {
            "Stop interactively one or more containers": "exegol stop",
            "Stop [blue]demo[/blue]": "exegol stop [blue]demo[/blue]"
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running stop module")
        return ExegolManager.stop


class Restart(Command, ContainerSelector, ContainerSpawnShell):
    """Restart an Exegol container"""

    def __init__(self) -> None:
        Command.__init__(self)
        ContainerSelector.__init__(self, self.groupArgs)
        ContainerSpawnShell.__init__(self, self.groupArgs)

        self._usages = {
            "Restart interactively one containers": "exegol restart",
            "Restart [blue]demo[/blue]": "exegol restart [blue]demo[/blue]"
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running restart module")
        return ExegolManager.restart


class Install(Command, ImageSelector):
    """Install or build Exegol image"""

    def __init__(self) -> None:
        Command.__init__(self)
        ImageSelector.__init__(self, self.groupArgs)

        # Create container build arguments
        self.build_profile = Option("build_profile",
                                    metavar="BUILD_PROFILE",
                                    nargs="?",
                                    action="store",
                                    help="Select the build profile used to create a local image.",
                                    completer=BuildProfileCompleter)
        self.build_log = Option("--build-log",
                                dest="build_log",
                                metavar="LOGFILE_PATH",
                                action="store",
                                help="Write image building logs to a file.")
        self.build_path = Option("--build-path",
                                 dest="build_path",
                                 metavar="DOCKERFILES_PATH",
                                 action="store",
                                 help=f"Path to the dockerfiles and sources.")

        # Create group parameter for container selection
        self.groupArgs.append(GroupArg({"arg": self.build_profile, "required": False},
                                       {"arg": self.build_log, "required": False},
                                       {"arg": self.build_path, "required": False},
                                       title="[bold cyan]Build[/bold cyan] [blue]specific options[/blue]"))

        self._usages = {
            "Install or build interactively an exegol image": "exegol install",
            "Install or update the [bright_blue]full[/bright_blue] image": "exegol install [bright_blue]full[/bright_blue]",
            "Build interactively a local image named [blue]myimage[/blue]": "exegol install [blue]myimage[/blue]",
            "Build the [blue]myimage[/blue] image based on the [bright_blue]full[/bright_blue] profile and log the operation": "exegol install [blue]myimage[/blue] [bright_blue]full[/bright_blue] --build-log /tmp/build.log",
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running install module")
        return ExegolManager.install


class Update(Command, ImageSelector):
    """Update an Exegol image"""

    def __init__(self) -> None:
        Command.__init__(self)
        ImageSelector.__init__(self, self.groupArgs)

        self.skip_git = Option("--skip-git",
                               dest="skip_git",
                               action="store_true",
                               help="Skip git updates (wrapper, image sources and exegol resources).")
        self.skip_images = Option("--skip-images",
                                  dest="skip_images",
                                  action="store_true",
                                  help="Skip images updates (exegol docker images).")

        # Create group parameter for container selection
        self.groupArgs.append(GroupArg({"arg": self.skip_git, "required": False},
                                       {"arg": self.skip_images, "required": False},
                                       title="[bold cyan]Update[/bold cyan] [blue]specific options[/blue]"))

        self._usages = {
            "Install or update interactively an exegol image": "exegol update",
            "Install or update the [bright_blue]full[/bright_blue] image": "exegol update [bright_blue]full[/bright_blue]"
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running update module")
        return ExegolManager.update


class Uninstall(Command, ImageMultiSelector):
    """Remove Exegol [default not bold]image(s)[/default not bold]"""

    def __init__(self) -> None:
        Command.__init__(self)
        ImageMultiSelector.__init__(self, self.groupArgs)

        self.force_mode = Option("-F", "--force",
                                 dest="force_mode",
                                 action="store_true",
                                 help="Remove image without interactive user confirmation.")

        # Create group parameter for container selection
        self.groupArgs.append(GroupArg({"arg": self.force_mode, "required": False},
                                       title="[bold cyan]Uninstall[/bold cyan] [blue]specific options[/blue]"))

        self._usages = {
            "Uninstall interactively one or more exegol images": "exegol uninstall",
            "Uninstall the [bright_blue]dev[/bright_blue] image": "exegol uninstall [bright_blue]dev[/bright_blue]"
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running uninstall module")
        return ExegolManager.uninstall


class Remove(Command, ContainerMultiSelector):
    """Remove Exegol [default not bold]container(s)[/default not bold]"""

    def __init__(self) -> None:
        Command.__init__(self)
        ContainerMultiSelector.__init__(self, self.groupArgs)

        self.force_mode = Option("-F", "--force",
                                 dest="force_mode",
                                 action="store_true",
                                 help="Remove container without interactive user confirmation.")

        # Create group parameter for container selection
        self.groupArgs.append(GroupArg({"arg": self.force_mode, "required": False},
                                       title="[bold cyan]Remove[/bold cyan] [blue]specific options[/blue]"))

        self._usages = {
            "Remove interactively one or more containers": "exegol remove",
            "Remove the [blue]demo[/blue] container": "exegol remove [blue]demo[/blue]"
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running remove module")
        return ExegolManager.remove


class Exec(Command, ContainerCreation, ContainerStart):
    """Execute a command on an Exegol container"""

    def __init__(self) -> None:
        Command.__init__(self)
        ContainerCreation.__init__(self, self.groupArgs)
        ContainerStart.__init__(self, self.groupArgs)

        # Overwrite default selectors
        for group in self.groupArgs.copy():
            # Find group containing default selector to remove them
            for parameter in group.options:
                if parameter.get('arg') == self.containertag or parameter.get('arg') == self.imagetag:
                    # Removing default GroupArg selector
                    self.groupArgs.remove(group)
                    break
        # Removing default selector objects
        self.containertag = None
        self.imagetag = None

        self.selector = Option("selector",
                               metavar="CONTAINER or IMAGE",
                               nargs='?',
                               action="store",
                               help="Tag used to target an Exegol container (by default) or an image (if --tmp is set).",
                               completer=HybridContainerImageCompleter)

        # Custom parameters
        self.exec = Option("exec",
                           metavar="COMMAND",
                           nargs="+",
                           action="store",
                           help="Execute a single command in the exegol container.",
                           completer=VoidCompleter)
        self.daemon = Option("-b", "--background",
                             action="store_true",
                             dest="daemon",
                             help="Executes the command in background as a daemon "
                                  "(default: [red not italic]False[/red not italic])")
        self.tmp = Option("--tmp",
                          action="store_true",
                          dest="tmp",
                          help="Creates a dedicated and temporary container to execute the command "
                               "(default: [red not italic]False[/red not italic])")

        # Create group parameter for container selection
        self.groupArgs.append(GroupArg({"arg": self.selector, "required": False},
                                       {"arg": self.exec, "required": False},
                                       {"arg": self.daemon, "required": False},
                                       {"arg": self.tmp, "required": False},
                                       title="[bold cyan]Exec[/bold cyan] [blue]specific options[/blue]"))

        self._usages = {
            "Execute the command [magenta]bloodhound[/magenta] in the container [blue]demo[/blue]":
                "exegol exec [blue]demo[/blue] [magenta]bloodhound[/magenta]",
            "Execute the command [magenta]'nmap -h'[/magenta] with console output":
                "exegol exec -v [blue]demo[/blue] [magenta]'nmap -h'[/magenta]",
            "Execute a command in [green]background[/green] within the [blue]demo[/blue] container":
                "exegol exec [green]-b[/green] [blue]demo[/blue] [magenta]bloodhound[/magenta]",
            "Execute the command [magenta]bloodhound[/magenta] in a temporary container based on the [bright_blue]full[/bright_blue] image":
                "exegol exec --tmp [bright_blue]full[/bright_blue] [magenta]bloodhound[/magenta]",
            "Execute a command in [green]background[/green] with a temporary container":
                "exegol exec [green]-b[/green] --tmp [bright_blue]full[/bright_blue] [magenta]bloodhound[/magenta]",
            "Execute the command [magenta]wireshark[/magenta] with [orange3]network admin[/orange3] privileged":
                "exegol exec [green]-b[/green] --tmp --disable-my-resources --cap [orange3]NET_ADMIN[/orange3] [bright_blue]full[/bright_blue] [magenta]wireshark[/magenta]",
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running exec module")
        return ExegolManager.exec


class Info(Command, ContainerSelector):
    """Show info on containers and images (local & remote)"""

    def __init__(self) -> None:
        Command.__init__(self)
        ContainerSelector.__init__(self, self.groupArgs)

        self._usages = {
            "Print containers and images essentials information": "exegol info",
            "Print the detailed configuration of the [blue]demo[/blue] container": "exegol info [blue]demo[/blue]",
            "Print verbose information": "exegol info [yellow3]-v[/yellow3]",
            "Print advanced information": "exegol info [yellow3]-vv[/yellow3]",
            "Print debug information": "exegol info [yellow3]-vvv[/yellow3]"
        }

    def __call__(self, *args, **kwargs):
        return ExegolManager.info


class Version(Command):
    """Print current Exegol version"""

    def __call__(self, *args, **kwargs):
        return lambda: None



import time
from datetime import datetime, timedelta
from typing import Optional

from docker.models.containers import Container

from exegol.utils.ExeLog import logger


class ContainerLogStream:

    def __init__(self, container: Container, start_date: Optional[datetime] = None, timeout: int = 5):
        # Container to extract logs from
        self.__container = container
        # Fetch more logs from this datetime
        self.__start_date: datetime = datetime.now() if start_date is None else start_date
        self.__since_date = self.__start_date
        self.__until_date: Optional[datetime] = None
        # The data stream is returned from the docker SDK. It can contain multiple line at the same.
        self.__data_stream = None
        self.__line_buffer = b''

        # Enable timeout if > 0. Passed timeout_date, the iterator will stop.
        self.__enable_timeout = timeout > 0
        self.__timeout_date: datetime = self.__since_date + timedelta(seconds=timeout)

        # Hint message flag
        self.__tips_sent = False
        self.__tips_timedelta = self.__start_date + timedelta(seconds=30)

    def __iter__(self):
        return self

    def __next__(self) -> str:
        """Get the next line of the stream"""
        if self.__until_date is None:
            self.__until_date = datetime.now()
        while True:
            # The data stream is fetch from the docker SDK once empty.
            if self.__data_stream is None:
                # The 'follow' mode cannot be used because there is no timeout mechanism and will stuck the process forever
                self.__data_stream = self.__container.logs(stream=True, follow=False, since=self.__since_date, until=self.__until_date)
            assert self.__data_stream is not None
            # Parsed the data stream to extract characters and merge them into a line.
            for streamed_char in self.__data_stream:
                self.__enable_timeout = False  # disable timeout if the container is up-to-date and support console logging
                # Add new char to the buffer + Unify \r\n to \n
                self.__line_buffer += streamed_char.replace(b'\r\n', b'\n')
                # When detecting an end of line, the buffer is returned as a single line.
                if b'\n' in self.__line_buffer:
                    lines = self.__line_buffer.split(b'\n')
                    self.__line_buffer = b'\n'.join(lines[1:]) if len(lines) > 1 else b''
                    if len(lines[0]) > 0:
                        return lines[0].decode('utf-8').strip()
            # When the data stream is empty, check if a timeout condition apply
            if self.__enable_timeout and self.__until_date >= self.__timeout_date:
                logger.debug("Container log stream timed-out")
                raise StopIteration
            elif not self.__tips_sent and self.__until_date >= self.__tips_timedelta:
                self.__tips_sent = True
                logger.info("Your start-up sequence takes time, your my-resource setup configuration may be significant.")
                logger.info("[orange3][Tips][/orange3] If you want to skip startup update, "
                            "you can use [green]CTRL+C[/green] and spawn a shell immediately. "
                            "[blue](Startup sequence will continue in background)[/blue]")
            # Prepare the next iteration to fetch next logs
            self.__data_stream = None
            self.__since_date = self.__until_date
            time.sleep(1)  # Wait for more logs
            self.__until_date = datetime.now()



import binascii
import logging
import os
from typing import Union, List, Tuple, Optional, cast, Sequence

from exegol.config.ConstantConfig import ConstantConfig
from exegol.config.EnvInfo import EnvInfo
from exegol.config.UserConfig import UserConfig
from exegol.console import ConsoleFormat
from exegol.console.ConsoleFormat import boolFormatter
from exegol.console.ExegolPrompt import Confirm
from exegol.console.TUI import ExegolTUI
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.console.cli.actions.GenericParameters import ContainerCreation
from exegol.exceptions.ExegolExceptions import ObjectNotFound, CancelOperation
from exegol.manager.UpdateManager import UpdateManager
from exegol.model.ContainerConfig import ContainerConfig
from exegol.model.ExegolContainer import ExegolContainer
from exegol.model.ExegolContainerTemplate import ExegolContainerTemplate
from exegol.model.ExegolImage import ExegolImage
from exegol.model.ExegolModules import ExegolModules
from exegol.model.SelectableInterface import SelectableInterface
from exegol.utils.DockerUtils import DockerUtils
from exegol.utils.ExeLog import logger, ExeLog


class ExegolManager:
    """Contains the main procedures of all actions available in Exegol"""

    # Cache data
    __container: Union[Optional[ExegolContainer], List[ExegolContainer]] = None
    __image: Union[Optional[ExegolImage], List[ExegolImage]] = None

    # Runtime default configuration
    __interactive_mode = False

    @classmethod
    def info(cls) -> None:
        """Print a list of available images and containers on the current host"""
        if logger.isEnabledFor(ExeLog.VERBOSE):
            logger.verbose("Listing user configurations")
            ExegolTUI.printTable(UserConfig().get_configs(), title="[not italic]:brain: [/not italic][gold3][g]User configurations[/g][/gold3]")
        if logger.isEnabledFor(ExeLog.ADVANCED):
            logger.verbose("Listing git repositories")
            ExegolTUI.printTable(UpdateManager.listGitStatus(), title="[not italic]:octopus: [/not italic][gold3][g]Project modules[/g][/gold3]")
        if bool(ParametersManager().containertag):
            # If the user have supplied a container name, show container config
            container = cls.__loadOrCreateContainer(ParametersManager().containertag, must_exist=True)
            if container is not None:
                assert type(container) is ExegolContainer
                ExegolTUI.printContainerRecap(container)
        else:
            # Without any parameter, show all images and containers info
            # Fetch data
            images = DockerUtils().listImages(include_version_tag=False)
            containers = DockerUtils().listContainers()
            # List and print images
            color = ConsoleFormat.getArchColor(ParametersManager().arch)
            logger.verbose(f"Listing local and remote Exegol images (filtering for architecture [{color}]{ParametersManager().arch}[/{color}])")
            ExegolTUI.printTable(images)
            # List and print containers
            logger.verbose("Listing local Exegol containers")
            logger.raw(f"[bold blue][*][/bold blue] Number of Exegol containers: {len(containers)}{os.linesep}",
                       markup=True)
            ExegolTUI.printTable(containers)

    @classmethod
    def start(cls) -> None:
        """Create and/or start an exegol container to finally spawn an interactive shell"""
        logger.info("Starting exegol")
        # Check if the first positional parameter have been supplied
        cls.__interactive_mode = not bool(ParametersManager().containertag)
        if not cls.__interactive_mode:
            logger.info("Arguments supplied with the command, skipping interactive mode")
        container = cls.__loadOrCreateContainer()
        assert container is not None and type(container) is ExegolContainer
        if not container.isNew():
            # Check and warn user if some parameters don't apply to the current session
            cls.__checkUselessParameters()
        container.start()
        container.spawnShell()

    @classmethod
    def exec(cls) -> None:
        """Create and/or start an exegol container to execute a specific command.
        The execution can be seen in console output or be relayed in the background as a daemon."""
        logger.info("Starting exegol")
        if ParametersManager().tmp:
            container = cls.__createTmpContainer(ParametersManager().selector)
            if not ParametersManager().daemon:
                container.exec(command=ParametersManager().exec, as_daemon=False, is_tmp=True)
                container.stop(timeout=2)
            else:
                # Command is passed at container creation in __createTmpContainer()
                logger.success(f"Command executed as entrypoint of the container {container.getDisplayName()}")
        else:
            container = cast(ExegolContainer, cls.__loadOrCreateContainer(override_container=ParametersManager().selector))
            container.exec(command=ParametersManager().exec, as_daemon=ParametersManager().daemon)

    @classmethod
    def stop(cls) -> None:
        """Stop an exegol container"""
        logger.info("Stopping exegol")
        container = cls.__loadOrCreateContainer(multiple=True, must_exist=True)
        assert container is not None and type(container) is list
        for c in container:
            c.stop(timeout=5)

    @classmethod
    def restart(cls) -> None:
        """Stop and start an exegol container"""
        container = cast(ExegolContainer, cls.__loadOrCreateContainer(must_exist=True))
        if container:
            container.stop(timeout=5)
            container.start()
            logger.success(f"Container [green]{container.name}[/green] successfully restarted!")
            container.spawnShell()

    @classmethod
    def install(cls) -> None:
        """Pull or build a docker exegol image"""
        try:
            if not ExegolModules().isExegolResourcesReady():
                raise CancelOperation
        except CancelOperation:
            # Error during installation, skipping operation
            logger.warning("Exegol resources have not been downloaded, the feature cannot be enabled")
        UpdateManager.updateImage(install_mode=True)

    @classmethod
    def update(cls) -> None:
        """Update python wrapper (git installation required) and Pull a docker exegol image"""
        if ParametersManager().offline_mode:
            logger.critical("It's not possible to update Exegol in offline mode. Please retry later with an internet connection.")
        if not ParametersManager().skip_git:
            UpdateManager.updateWrapper()
            UpdateManager.updateResources()
        if not ParametersManager().skip_images:
            UpdateManager.updateImage()

    @classmethod
    def uninstall(cls) -> None:
        """Remove an exegol image"""
        logger.info("Uninstalling an exegol image")
        # Set log level to verbose in order to show every image installed including the outdated.
        if not logger.isEnabledFor(ExeLog.VERBOSE):
            logger.setLevel(ExeLog.VERBOSE)
        images = cls.__loadOrInstallImage(multiple=True, must_exist=True)
        assert type(images) is list
        if len(images) == 0:
            return
        all_name = ", ".join([x.getName() for x in images])
        if not ParametersManager().force_mode and not Confirm(
                f"Are you sure you want to [red]permanently remove[/red] the following images? [orange3][ {all_name} ][/orange3]",
                default=False):
            logger.error("Aborting operation.")
            return
        for img in images:
            DockerUtils().removeImage(img)

    @classmethod
    def remove(cls) -> None:
        """Remove an exegol container"""
        logger.info("Removing an exegol container")
        containers = cls.__loadOrCreateContainer(multiple=True, must_exist=True)
        assert type(containers) is list
        if len(containers) == 0:
            logger.error("No containers were selected. Exiting.")
            return
        all_name = ", ".join([x.name for x in containers])
        if not ParametersManager().force_mode and not Confirm(
                f"Are you sure you want to [red]permanently remove[/red] the following containers? [orange3][ {all_name} ][/orange3]",
                default=False):
            logger.error("Aborting operation.")
            return
        for c in containers:
            c.remove()
            # If the image used is deprecated, it must be deleted after the removal of its container
            if c.image.isLocked() and UserConfig().auto_remove_images:
                DockerUtils().removeImage(c.image, upgrade_mode=True)

    @classmethod
    def print_version(cls) -> None:
        """Show exegol version (and context configuration on debug mode)"""

        logger.raw(f"[bold blue][*][/bold blue] Exegol is currently in version {UpdateManager.display_current_version()}{os.linesep}",
                   level=logging.INFO, markup=True)
        logger.raw(
            f"[bold magenta][*][/bold magenta] Exegol Discord serv.: [underline magenta]{ConstantConfig.discord}[/underline magenta]{os.linesep}",
            level=logging.INFO, markup=True)
        logger.raw(
            f"[bold magenta][*][/bold magenta] Exegol documentation: [underline magenta]{ConstantConfig.documentation}[/underline magenta]{os.linesep}",
            level=logging.INFO, markup=True)

        cls.print_sponsors()
        if 'a' in ConstantConfig.version:
            logger.empty_line()
            logger.warning("You are currently using an [red]Alpha[/red] version of Exegol, which may be unstable. "
                           "This version is a work in progress and bugs are expected.")
        elif 'b' in ConstantConfig.version:
            logger.empty_line()
            logger.warning("You are currently using a [orange3]Beta[/orange3] version of Exegol, which may be unstable.")

    @classmethod
    def print_debug_banner(cls) -> None:
        """Print header debug info"""
        logger.debug(f"Pip installation: {boolFormatter(ConstantConfig.pip_installed)}"
                     f"{'[bright_black](pipx)[/bright_black]' if ConstantConfig.pipx_installed else ''}")
        logger.debug(f"Git source installation: {boolFormatter(ConstantConfig.git_source_installation)}")
        logger.debug(f"Host OS: {EnvInfo.getHostOs().value} [bright_black]({EnvInfo.getDockerEngine().value})[/bright_black]")
        logger.debug(f"Arch: {EnvInfo.arch}")
        if EnvInfo.arch != EnvInfo.raw_arch:
            logger.debug(f"Raw arch: {EnvInfo.raw_arch}")
        if EnvInfo.isWindowsHost():
            logger.debug(f"Windows release: {EnvInfo.getWindowsRelease()}")
            logger.debug(f"Python environment: {EnvInfo.current_platform}")
            logger.debug(f"Docker engine: {EnvInfo.getDockerEngine().value}")
        logger.debug(f"Docker desktop: {boolFormatter(EnvInfo.isDockerDesktop())}")
        logger.debug(f"Shell type: {EnvInfo.getShellType()}")
        if UserConfig().auto_check_updates:
            UpdateManager.checkForWrapperUpdate()
        if UpdateManager.isUpdateAvailable():
            logger.empty_line()
            if Confirm(f"An [green]Exegol[/green] update is [orange3]available[/orange3] ({UpdateManager.display_current_version()} -> {UpdateManager.display_latest_version()}), do you want to update ?", default=True):
                UpdateManager.updateWrapper()
        else:
            logger.empty_line(log_level=logging.DEBUG)

    @classmethod
    def print_sponsors(cls) -> None:
        """Show exegol sponsors"""
        #logger.success("""We thank [link=https://www.capgemini.com/fr-fr/carrieres/offres-emploi/][blue]Capgemini[/blue][/link] for supporting the project [bright_black](helping with dev)[/bright_black] :pray:""")
        #logger.success("""We thank [link=https://www.hackthebox.com/][green]HackTheBox[/green][/link] for sponsoring the [bright_black]multi-arch[/bright_black] support :green_heart:""")
        pass

    @classmethod
    def __loadOrInstallImage(cls,
                             override_image: Optional[str] = None,
                             multiple: bool = False,
                             must_exist: bool = False) -> Union[Optional[ExegolImage], List[ExegolImage]]:
        """Select / Load (and install) an ExegolImage
        When must_exist is set to True, return None if no image are installed
        When multiple is set to True, return a list of ExegolImage
        Otherwise, always return an ExegolImage"""
        if cls.__image is not None:
            # Return cache
            return cls.__image
        image_tag = override_image if override_image is not None else ParametersManager().imagetag
        image_tags = ParametersManager().multiimagetag
        image_selection: Union[Optional[ExegolImage], List[ExegolImage]] = None
        # While an image have not been selected
        while image_selection is None:
            try:
                if image_tag is None and (image_tags is None or len(image_tags) == 0):
                    # Interactive (TUI) image selection
                    image_selection = cast(Union[Optional[ExegolImage], List[ExegolImage]],
                                           cls.__interactiveSelection(ExegolImage, multiple, must_exist))
                else:
                    # Select image by tag name (non-interactive)
                    if multiple:
                        image_selection = []
                        for image_tag in image_tags:
                            image_selection.append(DockerUtils().getInstalledImage(image_tag))
                    else:
                        image_selection = DockerUtils().getInstalledImage(image_tag)
            except ObjectNotFound:
                # ObjectNotFound is raised when the image_tag provided by the user does not match any existing image.
                if image_tag is not None:
                    logger.warning(f"The image named '{image_tag}' has not been found.")
                # If the user's selected image have not been found,
                # offer to build a local image with this name
                # (only if must_exist is not set)
                if not must_exist:
                    image_selection = UpdateManager.updateImage(image_tag)
                # Allow the user to interactively select another installed image
                image_tag = None
            except IndexError:
                # IndexError is raised when no image are available (not applicable when multiple is set, return an empty array)
                # (raised from TUI interactive selection)
                if must_exist:
                    # If there is no image installed, return none
                    logger.error("Nothing to do.")
                    return [] if multiple else None
                elif image_tag is not None:
                    # If the user's selected image have not been found, offer the choice to build a local image at this name
                    # (only if must_exist is not set)
                    image_selection = UpdateManager.updateImage(image_tag)
                    image_tag = None
                else:
                    logger.critical("No image are installed or available, check your internet connection and install an image with the command [green]exegol install[/green].")
            # Checks if an image has been selected
            if image_selection is None:
                # If not, retry the selection
                logger.error("No image has been selected.")
                continue

            # Check if every image are installed
            install_status, checked_images = cls.__checkImageInstallationStatus(image_selection, multiple, must_exist)
            if not install_status:
                # If one of the image is not install where it supposed to, restart the selection
                # allowing him to interactively choose another image
                image_selection, image_tag = None, None
                continue

            cls.__image = cast(Union[Optional[ExegolImage], List[ExegolImage]], checked_images)
        return cls.__image

    @classmethod
    def __checkImageInstallationStatus(cls,
                                       image_selection: Union[ExegolImage, List[ExegolImage]],
                                       multiple: bool = False,
                                       must_exist: bool = False
                                       ) -> Tuple[bool, Optional[Union[ExegolImage, ExegolContainer, List[ExegolImage], List[ExegolContainer]]]]:
        """Checks if the selected images are installed and ready for use.
        returns false if the images are supposed to be already installed."""
        # Checks if one or more images have been selected and unifies the format into a list.
        reverse_type = False
        check_img: List[ExegolImage]
        if type(image_selection) is ExegolImage:
            check_img = [image_selection]
            # Tag of the operation to reverse it before the return
            reverse_type = True
        elif type(image_selection) is list:
            check_img = image_selection
        else:
            check_img = []

        # Check if every image are installed
        for i in range(len(check_img)):
            if not check_img[i].isInstall():
                # Is must_exist is set, every image are supposed to be already installed
                if must_exist:
                    logger.error(f"The selected image '{check_img[i].getName()}' is not installed.")
                    # If one of the image is not install, return False to restart the selection
                    return False, None
                else:
                    # Check if the selected image is installed and install it
                    logger.warning("The selected image is not installed.")
                    # Download remote image
                    if DockerUtils().downloadImage(check_img[i], install_mode=True):
                        # Select installed image
                        check_img[i] = DockerUtils().getInstalledImage(check_img[i].getName())
                    else:
                        logger.error("This image cannot be installed.")
                        return False, None

        if reverse_type and not multiple:
            # Restoration of the original type
            return True, check_img[0]
        return True, check_img

    @classmethod
    def __loadOrCreateContainer(cls,
                                override_container: Optional[str] = None,
                                multiple: bool = False,
                                must_exist: bool = False) -> Union[Optional[ExegolContainer], List[ExegolContainer]]:
        """Select one or more ExegolContainer
        Or create a new ExegolContainer if no one already exist (and must_exist is not set)
        When must_exist is set to True, return None if no container exist
        When multiple is set to True, return a list of ExegolContainer"""
        if cls.__container is not None:
            # Return cache
            return cls.__container
        container_tag: Optional[str] = override_container if override_container is not None else ParametersManager().containertag
        container_tags: Optional[List[str]] = None
        if ParametersManager().multicontainertag:
            container_tags = []
            for tag in ParametersManager().multicontainertag:
                # Prevent duplicate tag selection
                if tag not in container_tags:
                    container_tags.append(tag)
        try:
            if container_tag is None and (container_tags is None or len(container_tags) == 0):
                # Interactive container selection
                cls.__container = cast(Union[Optional[ExegolContainer], List[ExegolContainer]],
                                       cls.__interactiveSelection(ExegolContainer, multiple, must_exist))
            else:
                # Try to find the corresponding container
                if multiple:
                    cls.__container = []
                    assert container_tags is not None
                    # test each user tag
                    for container_tag in container_tags:
                        try:
                            cls.__container.append(DockerUtils().getContainer(container_tag))
                        except ObjectNotFound:
                            # on multi select, an object not found is not critical
                            if must_exist:
                                # If the selected tag doesn't match any container, print an alert and continue
                                logger.warning(f"The container named '{container_tag}' has not been found")
                            else:
                                # If there is a multi select without must_exist flag, raise an error
                                # because multi container creation is not supported
                                raise NotImplementedError
                else:
                    assert container_tag is not None
                    cls.__container = DockerUtils().getContainer(container_tag)
        except (ObjectNotFound, IndexError):
            # ObjectNotFound is raised when the container_tag provided by the user does not match any existing container.
            # IndexError is raise when no container exist (raised from TUI interactive selection)
            # Create container
            if must_exist:
                logger.warning(f"The container named '{container_tag}' has not been found")
                return [] if multiple else None
            return cls.__createContainer(container_tag)
        assert cls.__container is not None
        return cast(Union[Optional[ExegolContainer], List[ExegolContainer]], cls.__container)

    @classmethod
    def __interactiveSelection(cls,
                               object_type: type,
                               multiple: bool = False,
                               must_exist: bool = False) -> \
            Union[ExegolImage, ExegolContainer, Sequence[ExegolImage], Sequence[ExegolContainer]]:
        """Interactive object selection process, depending on object_type.
        object_type can be ExegolImage or ExegolContainer."""
        object_list: Sequence[SelectableInterface]
        # Object listing depending on the type
        if object_type is ExegolContainer:
            # List all images available
            object_list = DockerUtils().listContainers()
        elif object_type is ExegolImage:
            # List all images available
            object_list = DockerUtils().listInstalledImages() if must_exist else DockerUtils().listImages()
        else:
            logger.critical("Unknown object type during interactive selection. Exiting.")
            raise Exception
        # Interactive choice with TUI
        user_selection: Union[SelectableInterface, Sequence[SelectableInterface], str]
        if multiple:
            user_selection = ExegolTUI.multipleSelectFromTable(object_list, object_type=object_type)
        else:
            user_selection = ExegolTUI.selectFromTable(object_list, object_type=object_type, allow_None=not must_exist)
            # Check if the user has chosen an existing object
            if type(user_selection) is str:
                # Otherwise, create a new object with the supplied name
                if object_type is ExegolContainer:
                    user_selection = cls.__createContainer(user_selection)
                else:
                    # Calling buildAndLoad directly, no need to ask confirmation, already done by TUI.
                    user_selection = UpdateManager.buildAndLoad(user_selection)
        return cast(Union[ExegolImage, ExegolContainer, List[ExegolImage], List[ExegolContainer]], user_selection)

    @classmethod
    def __prepareContainerConfig(cls) -> ContainerConfig:
        """Create Exegol configuration with user input"""
        try:
            # Create default exegol config
            config = ContainerConfig()
            # Container configuration from user CLI options
            if ParametersManager().X11:
                config.enableGUI()
            if ParametersManager().share_timezone:
                config.enableSharedTimezone()
            config.setNetworkMode(ParametersManager().host_network)
            if ParametersManager().ports is not None:
                for port in ParametersManager().ports:
                    config.addRawPort(port)
            if ParametersManager().my_resources:
                config.enableMyResources()
            if ParametersManager().exegol_resources:
                config.enableExegolResources()
            if ParametersManager().log:
                config.enableShellLogging(ParametersManager().log_method,
                                          UserConfig().shell_logging_compress ^ ParametersManager().log_compress)
            if ParametersManager().workspace_path:
                if ParametersManager().mount_current_dir:
                    logger.warning(f'Workspace conflict detected (-cwd cannot be use with -w). Using: {ParametersManager().workspace_path}')
                config.setWorkspaceShare(ParametersManager().workspace_path)
            elif ParametersManager().mount_current_dir:
                config.enableCwdShare()
            if ParametersManager().privileged:
                config.setPrivileged()
            elif ParametersManager().capabilities is not None:
                for cap in ParametersManager().capabilities:
                    config.addCapability(cap)
            if ParametersManager().volumes is not None:
                for volume in ParametersManager().volumes:
                    config.addRawVolume(volume)
            if ParametersManager().devices is not None:
                for device in ParametersManager().devices:
                    config.addUserDevice(device)
            if ParametersManager().vpn is not None:
                config.enableVPN()
            if ParametersManager().envs is not None:
                for env in ParametersManager().envs:
                    config.addRawEnv(env)
            if (UserConfig().desktop_default_enable ^ ParametersManager().desktop) or ParametersManager().desktop_config != "":
                config.enableDesktop(ParametersManager().desktop_config)
            if ParametersManager().comment:
                config.addComment(ParametersManager().comment)
            return config
        except CancelOperation as e:
            logger.critical(f"Unable to create a new container: {e}")
            raise e

    @classmethod
    def __createContainer(cls, name: Optional[str]) -> ExegolContainer:
        """Create an ExegolContainer"""
        logger.verbose("Configuring new exegol container")
        # Create exegol config
        image: Optional[ExegolImage] = cast(ExegolImage, cls.__loadOrInstallImage())
        config = cls.__prepareContainerConfig()
        assert image is not None  # load or install return an image
        model = ExegolContainerTemplate(name, config, image, hostname=ParametersManager().hostname)

        # Recap
        ExegolTUI.printContainerRecap(model)
        if cls.__interactive_mode:
            if not model.image.isUpToDate() and \
                    Confirm("Do you want to [green]update[/green] the selected image?", False):
                image = UpdateManager.updateImage(model.image.getName())
                if image is not None:
                    model.image = image
                    ExegolTUI.printContainerRecap(model)
            command_options = []
            while not Confirm("Is the container configuration [green]correct[/green]?", default=True):
                command_options = model.config.interactiveConfig(model.name)
                ExegolTUI.printContainerRecap(model)
            logger.info(f"Command line of the configuration: "
                        f"[green]exegol start {model.name} {model.image.getName()} {' '.join(command_options)}[/green]")
            logger.info("To use exegol [orange3]without interaction[/orange3], "
                        "read CLI options with [green]exegol start -h[/green]")

        container = DockerUtils().createContainer(model)
        container.postCreateSetup()
        return container

    @classmethod
    def __createTmpContainer(cls, image_name: Optional[str] = None) -> ExegolContainer:
        """Create a temporary ExegolContainer with custom entrypoint"""
        logger.verbose("Configuring new exegol container")
        # Create exegol config
        config = cls.__prepareContainerConfig()
        # When container exec a command as a daemon, the execution must be set on the container's entrypoint
        if ParametersManager().daemon:
            # Using formatShellCommand to support zsh aliases
            exec_payload, str_cmd = ExegolContainer.formatShellCommand(ParametersManager().exec, entrypoint_mode=True)
            config.entrypointRunCmd()
            config.addEnv("CMD", str_cmd)
            config.addEnv("DISABLE_AUTO_UPDATE", "true")
        # Workspace must be disabled for temporary container because host directory is never deleted
        config.disableDefaultWorkspace()
        name = f"tmp-{binascii.b2a_hex(os.urandom(4)).decode('ascii')}"
        image: ExegolImage = cast(ExegolImage, cls.__loadOrInstallImage(override_image=image_name))
        model = ExegolContainerTemplate(name, config, image, hostname=ParametersManager().hostname)

        # Mount entrypoint as a volume (because in tmp mode the container is created with run instead of create method)
        model.config.addVolume(ConstantConfig.entrypoint_context_path_obj, "/.exegol/entrypoint.sh", must_exist=True, read_only=True)

        container = DockerUtils().createContainer(model, temporary=True)
        container.postCreateSetup(is_temporary=True)
        return container

    @classmethod
    def __checkUselessParameters(cls) -> None:
        """Checks if the container creation parameters have not been filled in when the container already existed"""
        # Get defaults parameters
        creation_parameters = ContainerCreation([]).__dict__
        # Get parameters from user input
        user_inputs = ParametersManager().parameters.__dict__
        detected = []
        for param in creation_parameters.keys():
            # Skip parameters useful in a start context
            if param in ('containertag',):
                continue
            # For each parameter, check if it's not None and different from the default
            current_option = creation_parameters.get(param)
            if user_inputs.get(param) is not None and current_option is not None and \
                    user_inputs.get(param) != current_option.kwargs.get('default'):
                # If the supplied parameter is positional, getting his printed name
                name = current_option.kwargs.get('metavar')
                if not name:
                    # if not, using the args name
                    detected.append(' / '.join(current_option.args))
                else:
                    detected.append(name)
        if len(detected) > 0:
            logger.warning(f"These parameters ({', '.join(detected)}) have been entered although the container already "
                           f"exists, they will not be taken into account.")



from pathlib import Path

__version__ = "4.3.10"


class ConstantConfig:
    """Constant parameters information"""
    # Exegol Version
    version: str = __version__

    # Exegol documentation link
    documentation: str = "https://exegol.rtfd.io/"
    discord: str = "https://discord.gg/cXThyp7D6P"
    # OS Dir full root path of exegol project
    src_root_path_obj: Path = Path(__file__).parent.parent.parent.resolve()
    # Path of the entrypoint.sh
    entrypoint_context_path_obj: Path = src_root_path_obj / "exegol/utils/imgsync/entrypoint.sh"
    # Path of the spawn.sh
    spawn_context_path_obj: Path = src_root_path_obj / "exegol/utils/imgsync/spawn.sh"
    # Exegol config directory
    exegol_config_path: Path = Path().home() / ".exegol"
    # Docker Desktop for mac config file
    docker_desktop_mac_config_path = Path().home() / "Library/Group Containers/group.com.docker"
    docker_desktop_windows_config_short_path = "AppData/Roaming/Docker"
    docker_desktop_windows_config_path = Path().home() / docker_desktop_windows_config_short_path
    # Install mode, check if Exegol has been git cloned or installed using pip package
    git_source_installation: bool = (src_root_path_obj / '.git').is_dir()
    pip_installed: bool = src_root_path_obj.name == "site-packages"
    pipx_installed: bool = "/pipx/venvs/" in src_root_path_obj.as_posix()
    # Dockerhub Exegol images repository
    DOCKER_HUB: str = "hub.docker.com"  # Don't handle docker login operations
    DOCKER_REGISTRY: str = "registry-1.docker.io"  # Don't handle docker login operations
    IMAGE_NAME: str = "nwodtuhs/exegol"
    GITHUB_REPO: str = "ThePorgs/Exegol"
    # Docker volume names (no docker volume used at this moment)
    # Resources repository
    EXEGOL_IMAGES_REPO: str = "https://github.com/ThePorgs/Exegol-images.git"
    EXEGOL_RESOURCES_REPO: str = "https://github.com/ThePorgs/Exegol-resources.git"



# Exceptions specific to the successful operation of exegol
class ObjectNotFound(Exception):
    """Custom exception when a specific container do not exist"""
    pass


class ProtocolNotSupported(Exception):
    """Custom exception when a specific network protocol is not supported"""
    pass


class CancelOperation(Exception):
    """Custom exception when an error occurred and the operation must be canceled ou skipped"""
    pass



import argparse
from logging import CRITICAL
from typing import Optional, List, Union, Dict, cast

import argcomplete

from exegol.console.cli.actions.Command import Command, Option
from exegol.utils.ExeLog import logger


class ExegolArgParse(argparse.ArgumentParser):
    """Overloading of the main parsing (argparse.ArgumentParser) class"""

    # Using Exelog to print built-in parser message
    def _print_message(self, message: str, file=None) -> None:
        if message:
            logger.raw(message, level=CRITICAL, rich_parsing=True)


class Parser:
    """Custom Exegol CLI Parser. Main controller of argument building and parsing."""

    __description = """This Python script is a wrapper for Exegol. It can be used to easily manage Exegol on your machine.

[bold magenta]Exegol documentation:[/bold magenta] [underline magenta]https://exegol.rtfd.io[/underline magenta]"""
    __formatter_class = argparse.RawTextHelpFormatter

    def __init__(self, actions: List[Command]):
        """Custom parser creation"""
        # Defines every actions available
        self.__actions: List[Command] = actions
        # Create & init root parser
        self.__root_parser: ExegolArgParse
        self.__init_parser()
        # Configure root options
        # (WARNING: these parameters are duplicate with every sub-parse, cannot use positional args here)
        self.__set_options(self.__root_parser, Command())  # Add global arguments from Command to the root parser
        # Create & fill sub-parser
        self.subParser = self.__root_parser.add_subparsers(help="Description of the actions")
        self.__set_action_parser()

    def __init_parser(self) -> None:
        """Root parser creation"""

        self.__root_parser = ExegolArgParse(
            description=self.__description,
            epilog=Command().formatEpilog(),
            formatter_class=self.__formatter_class,
        )

    def __set_action_parser(self) -> None:
        """Create sub-parser for each action and configure it"""
        self.__root_parser._positionals.title = "[green]Required arguments[/green]"
        for action in self.__actions:
            # Each action has a dedicated sub-parser with different options
            # the 'help' description of the current action is retrieved
            # from the comment of the corresponding action class
            if action.__doc__ is None:
                action.__doc__ = "Unknown action"
            sub_parser = self.subParser.add_parser(action.name, help=action.__doc__,
                                                   description=action.__doc__ + f"""\n
[bold magenta]Exegol documentation:[/bold magenta] [underline magenta]https://exegol.rtfd.io/en/latest/exegol-wrapper/{action.name}.html[/underline magenta]""",
                                                   epilog=action.formatEpilog(),
                                                   formatter_class=self.__formatter_class)
            sub_parser.set_defaults(action=action)
            self.__set_options(sub_parser, target=action)

    def __set_options(self, sub_parser: argparse.ArgumentParser, target: Optional[Command] = None) -> None:
        """Add different groups and parameters/options in the current sub_parser"""
        global_set = False  # Only one group can be global at the time
        # Load actions to be processed (default: every action from cls)
        actions_list = [target] if target else self.__actions
        for action in actions_list:
            # On each action, fetch every group to be processed
            for argument_group in action.groupArgs:
                group_parser: argparse._ActionsContainer
                if argument_group.is_global and not global_set:
                    # If the current group is global (ex: 'Optional arguments'),
                    # overwriting parser main group before adding custom parameters
                    global_set = True  # The setup application should be run only once
                    sub_parser._optionals.title = argument_group.title  # Overwriting default argparse title
                    group_parser = sub_parser  # Subparser is directly used to add arguments
                else:
                    # In every other case, a dedicated group is created in the parser
                    group_parser = sub_parser.add_argument_group(argument_group.title,
                                                                 description=argument_group.description)
                # once the group is created in the parser, the arguments can be added to it
                option: Dict[str, Union[Option, bool]]
                for option in argument_group.options:
                    # Retrieve Option object from the Dict
                    assert type(option["arg"]) is Option
                    argument = cast(Option, option["arg"])
                    # Pop is required here to removed unknown parameter from the action object before argparse
                    completer = argument.kwargs.pop("completer", None)
                    try:
                        arg = group_parser.add_argument(*argument.args, **argument.kwargs)
                    except argparse.ArgumentError:
                        continue
                    # Add argument with its config to the parser
                    if completer is not None:
                        arg.completer = completer  # type: ignore

    def run_parser(self) -> argparse.Namespace:
        """Execute argparse to retrieve user options from argv"""
        argcomplete.autocomplete(self.__root_parser)
        return self.__root_parser.parse_args()

    def print_help(self) -> None:
        """Force argparse to display the help message"""
        self.__root_parser.print_help()



class SelectableInterface:
    """Generic class used to select objects in the user TUI"""

    def getKey(self) -> str:
        """Universal unique key getter"""
        raise NotImplementedError

    def __eq__(self, other) -> bool:
        """Generic '==' operator overriding matching object key"""
        return other == self.getKey()



import os
import re
from typing import Union, Optional, List, Dict, Type, Generator, Set, cast, Sequence, Tuple

from rich import box
from rich.progress import TextColumn, BarColumn, TransferSpeedColumn, TimeElapsedColumn, TimeRemainingColumn, TaskID
from rich.prompt import Prompt
from rich.table import Table

from exegol.config.EnvInfo import EnvInfo
from exegol.console import ConsoleFormat
from exegol.console.ConsoleFormat import boolFormatter, getColor, richLen
from exegol.console.ExegolProgress import ExegolProgress
from exegol.console.ExegolPrompt import Confirm
from exegol.console.LayerTextColumn import LayerTextColumn
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.model.ExegolContainer import ExegolContainer
from exegol.model.ExegolContainerTemplate import ExegolContainerTemplate
from exegol.model.ExegolImage import ExegolImage
from exegol.model.SelectableInterface import SelectableInterface
from exegol.utils.ExeLog import logger, console, ExeLog


class ExegolTUI:
    """Class gathering different methods of Terminal User Interface (or TUI)"""

    @staticmethod
    def downloadDockerLayer(stream: Generator, quick_exit: bool = False) -> None:
        """Rich interface for docker image layer download from SDK stream"""
        layers: Set[str] = set()
        layers_downloaded: Set[str] = set()
        layers_extracted: Set[str] = set()
        downloading: Dict[str, TaskID] = {}
        extracting: Dict[str, TaskID] = {}
        # Create progress bar with columns
        with ExegolProgress(TextColumn("{task.description}", justify="left"),
                            BarColumn(bar_width=None),
                            "[progress.percentage]{task.percentage:>3.1f}%",
                            "•",
                            LayerTextColumn("[bold]{task.completed}/{task.total}", "layer"),
                            "•",
                            TransferSpeedColumn(),
                            "•",
                            TimeElapsedColumn(),
                            "•",
                            TimeRemainingColumn(),
                            transient=True,
                            console=console) as progress:
            task_layers_download = progress.add_task("[bold red]Downloading layers...", total=0)
            task_layers_extract = progress.add_task("[bold gold1]Extracting layers...", total=0, start=False)
            for line in stream:  # Receiving stream from docker API
                status = line.get("status", '')
                error = line.get("error", '')
                layer_id = line.get("id")
                if error != "":
                    logger.error(f"Docker download error: {error}")
                    logger.critical(f"An error occurred during the image download. Exiting.")
                if status == "Pulling fs layer":  # Identify new layer to download
                    layers.add(layer_id)
                    progress.update(task_layers_download, total=len(layers))
                    progress.update(task_layers_extract, total=len(layers))
                elif "Pulling from " in status:  # Rename task with image name
                    progress.getTask(task_layers_download).description = \
                        f"[bold red]Downloading {status.replace('Pulling from ', '')}:{line.get('id', 'latest')}"
                    progress.getTask(task_layers_extract).description = \
                        f"[bold gold1]Extracting {status.replace('Pulling from ', '')}:{line.get('id', 'latest')}"
                elif status == "Download complete" or status == "Pull complete":  # Mark task as complete and remove it from the pool
                    # Select task / layer pool depending on the status
                    task_pool = downloading
                    layer_pool = layers_downloaded
                    if status == "Pull complete":
                        task_pool = extracting
                        layer_pool = layers_extracted
                    # Tagging current layer as ended
                    layer_pool.add(layer_id)
                    # Remove finished layer progress bar
                    layer_task = task_pool.get(layer_id)
                    if layer_task is not None:
                        progress.remove_task(layer_task)  # Remove progress bar
                        task_pool.pop(layer_id)  # Remove task from pool
                    # Update global task completion status
                    progress.update(task_layers_download, completed=len(layers_downloaded))
                    progress.update(task_layers_extract, completed=len(layers_extracted))
                elif status == "Downloading" or status == "Extracting":  # Handle download or extract progress
                    task_pool = downloading
                    if status == "Extracting":
                        task_pool = extracting
                        if not progress.getTask(task_layers_extract).started:
                            progress.start_task(task_layers_extract)
                    task_id = task_pool.get(layer_id)
                    progressDetail = line.get("progressDetail", {})
                    if task_id is None:  # If this is a new layer, create a new task accordingly
                        task_id = progress.add_task(
                            f"[{'blue' if status == 'Downloading' else 'magenta'}]{status} {layer_id}",
                            total=progressDetail.get("total", 100),
                            layer=layer_id)
                        task_pool[layer_id] = task_id
                    # Updating task progress
                    progress.update(task_id, completed=progressDetail.get("current", 100))
                    if status == "Extracting" and progressDetail.get("current", 0) == progressDetail.get("total", 100):
                        progress.update(task_id, description=f"[green]Checksum {layer_id} ...")
                elif "Image is up to date" in status or "Status: Downloaded newer image for" in status:
                    logger.success(status)
                    # When 'quick_exit' is set, breaking the download TUI progress to handle the next part of the same stream logs
                    if quick_exit:
                        break
                elif status in ["Waiting", "Verifying Checksum"]:
                    # Ignore these status messages
                    continue
                elif status == "Already exists":
                    # Layers that already exists can be added as fully completed without creating a specific task
                    layers.add(layer_id)
                    layers_downloaded.add(layer_id)
                    layers_extracted.add(layer_id)
                else:
                    logger.debug(line)

    @staticmethod
    def buildDockerImage(build_stream: Generator) -> None:
        """Rich interface for docker image building from SDK stream"""
        # Prepare log file
        logfile = None
        if ParametersManager().build_log is not None:
            # Opening log file in line buffering mode (1) to support tail -f [file]
            logfile = open(ParametersManager().build_log, 'a', buffering=1)
        # Follow stream
        for line in build_stream:
            stream_text = line.get("stream", '')
            error_text = line.get("error", '')
            if logfile is not None:
                logfile.write(stream_text)
                logfile.write(error_text)
            if error_text != "":
                logger.error(f"Docker build error: {error_text}")
                logger.critical(
                    f"An error occurred during the image build (code: {line.get('errorDetail', {}).get('code', '?')}). Exiting.")
            if stream_text.strip() != '':
                if "Step" in stream_text:
                    logger.info(stream_text.rstrip())
                elif "--->" in stream_text or \
                        "Removing intermediate container " in stream_text or \
                        re.match(r"Successfully built [a-z0-9]{12}", stream_text) or \
                        re.match(r"^Successfully tagged ", stream_text):
                    logger.verbose(stream_text.rstrip())
                else:
                    logger.raw(stream_text, level=ExeLog.ADVANCED)
            if ': FROM ' in stream_text:
                logger.info("Downloading base image")
                ExegolTUI.downloadDockerLayer(build_stream, quick_exit=True)
        if logfile is not None:
            logfile.close()

    @staticmethod
    def printTable(data: Union[Sequence[SelectableInterface], Sequence[str], Sequence[Dict[str, str]]],
                   title: Optional[str] = None,
                   safe_key: bool = False) -> None:
        """Printing Rich table for a list of object.
        Set safe_key to override the key selection"""
        logger.empty_line()
        table = Table(title=title, show_header=True, header_style="bold blue", border_style="grey35",
                      box=box.SQUARE, title_justify="left")
        if len(data) == 0:
            logger.debug("No data supplied")
            return
        else:
            if type(data[0]) is ExegolImage:
                ExegolTUI.__buildImageTable(table, cast(Sequence[ExegolImage], data), safe_key=safe_key)
            elif type(data[0]) is ExegolContainer:
                ExegolTUI.__buildContainerTable(table, cast(Sequence[ExegolContainer], data), safe_key=safe_key)
            elif type(data[0]) is str:
                if title is not None:
                    ExegolTUI.__buildStringTable(table, cast(Sequence[str], data), cast(str, title))
                else:
                    ExegolTUI.__buildStringTable(table, cast(Sequence[str], data))
            elif type(data[0]) is dict:
                ExegolTUI.__buildDictTable(table, cast(Sequence[Dict[str, str]], data))
            else:
                logger.error(f"Print table of {type(data[0])} is not implemented")
                raise NotImplementedError
        console.print(table)
        logger.empty_line()

    @staticmethod
    def __buildImageTable(table: Table, data: Sequence[ExegolImage], safe_key: bool = False) -> None:
        """Building Rich table from a list of ExegolImage"""
        table.title = "[not italic]:flying_saucer: [/not italic][gold3][g]Available images[/g][/gold3]"
        # Define columns
        verbose_mode = logger.isEnabledFor(ExeLog.VERBOSE)
        debug_mode = logger.isEnabledFor(ExeLog.ADVANCED)
        if safe_key:
            table.add_column("Option")
        if verbose_mode:
            table.add_column("Id")
        table.add_column("Image tag")
        if verbose_mode:
            table.add_column("Download size")
            table.add_column("Size on disk")
            table.add_column("Build date (UTC)")
        else:
            # Depending on whether the image has already been downloaded or not,
            # it will show the download size or the size on disk
            table.add_column("Size")
        table.add_column("Status")
        # Load data into the table
        for i in range(len(data)):
            image = data[i]
            if verbose_mode:
                if safe_key:
                    table.add_row(str(i + 1), image.getLocalId(), image.getDisplayName(), image.getDownloadSize(),
                                  image.getRealSize(), image.getBuildDate(), image.getStatus())
                else:
                    table.add_row(image.getLocalId(), image.getDisplayName(), image.getDownloadSize(),
                                  image.getRealSize(), image.getBuildDate(), image.getStatus())
            else:
                if safe_key:
                    table.add_row(str(i + 1), image.getDisplayName(), image.getRealSize(), image.getStatus())
                else:
                    table.add_row(image.getDisplayName(), image.getRealSize(), image.getStatus())

    @staticmethod
    def __buildContainerTable(table: Table, data: Sequence[ExegolContainer], safe_key: bool = False) -> None:
        """Building Rich table from a list of ExegolContainer"""
        table.title = "[not italic]:alien: [/not italic][gold3][g]Available containers[/g][/gold3]"
        # Define columns
        verbose_mode = logger.isEnabledFor(ExeLog.VERBOSE)
        debug_mode = logger.isEnabledFor(ExeLog.ADVANCED)
        if verbose_mode:
            table.add_column("Id")
        table.add_column("Container tag")
        table.add_column("State")
        table.add_column("Image tag")
        table.add_column("Configurations")
        if verbose_mode:
            table.add_column("Mounts")
            table.add_column("Devices")
            table.add_column("Ports")
            table.add_column("Envs")
        # Load data into the table
        for container in data:
            if verbose_mode:
                table.add_row(container.getId(), container.getDisplayName(), container.getTextStatus(), container.image.getDisplayName(),
                              container.config.getTextFeatures(verbose_mode),
                              container.config.getTextMounts(debug_mode),
                              container.config.getTextDevices(debug_mode),
                              container.config.getTextPorts(),
                              container.config.getTextEnvs(debug_mode))
            else:
                table.add_row(container.getDisplayName(), container.getTextStatus(), container.image.getDisplayName(),
                              container.config.getTextFeatures(verbose_mode))

    @staticmethod
    def __buildStringTable(table: Table, data: Sequence[str], title: str = "Key") -> None:
        """Building a simple Rich table from a list of string"""
        table.title = title
        table.min_width = richLen(title)
        # Define columns
        table.add_column(title)
        table.show_header = False
        # Load data into the table
        for string in data:
            table.add_row(string)

    @staticmethod
    def __buildDictTable(table: Table, data_array: Sequence[Dict[str, str]]) -> None:
        """Building a simple Rich table from a list of string"""
        # Define columns from dict keys
        for column in data_array[0].keys():
            table.add_column(column.capitalize())
        # Load data into the table
        for data in data_array:
            # Array is directly pass as *args to handle dynamic columns number
            table.add_row(*data.values())

    @classmethod
    def selectFromTable(cls,
                        data: Sequence[SelectableInterface],
                        object_type: Optional[Type] = None,
                        default: Optional[str] = None,
                        allow_None: bool = False,
                        conflict_mode: bool = False) -> Union[SelectableInterface, str]:
        """Return an object (implementing SelectableInterface) selected by the user
        Return a str when allow_none is true and no object have been selected
        Raise IndexError of the data list is empty.
        Set conflict_mode to override the key in order to select a specific object with duplicate name"""
        cls.__isInteractionAllowed()
        # Check if there is at least one object in the list
        if len(data) == 0:
            if object_type is ExegolImage:
                logger.warning("No images are available for selection")
            elif object_type is ExegolContainer:
                logger.warning("No containers have been created yet")
            else:
                # Using container syntax by default
                logger.warning("No containers have been created yet")
            raise IndexError
        object_type = type(data[0])
        object_name = "container" if object_type is ExegolContainer else "image"
        action = "create" if object_type is ExegolContainer else "build"
        # Get a list of every choice available
        choices: List[str] = [obj.getKey() for obj in data]
        if conflict_mode or (len(data) > 1 and len(set(choices)) == 1):
            conflict_mode = True
            choices = [str(k) for k in range(1, len(data) + 1)]
        # Print data list
        cls.printTable(data, safe_key=conflict_mode)
        # If no default have been supplied, using the first one
        if default is None:
            default = choices[0]
        # When allow_none is enabled, disabling choices restriction
        if allow_None:
            choices_select: Optional[List[str]] = None
            logger.info(
                f"You can use a name that does not already exist to {action} a new {object_name}"
                f"{' from local sources' if object_type is ExegolImage else ''}")
        else:
            choices_select = choices
        while True:
            match = []
            choice = Prompt.ask(
                f"[bold blue][?][/bold blue] Select {'an' if object_type is ExegolImage else 'a'} {object_name} by its name",
                default=default, choices=choices_select,
                show_choices=False)
            if conflict_mode:
                # In conflict mode, choice are only index number offset by 1
                return data[int(choice)-1]
            for option in data:
                if choice == option:
                    match.append(option)
            if len(match) == 1:
                return match[0]
            elif len(match) > 1:
                logger.error(f"Conflict detected ! Multiple {object_name} have the same name, please select the intended one.")
                return cls.selectFromTable(match, object_type, default=None, allow_None=False, conflict_mode=True)
            if allow_None:
                if Confirm(
                        f"No {object_name} is available under this name, do you want to {action} it?",
                        default=True):
                    return choice
                logger.info(f"[red]Please select one of the available {object_name}s[/red]")
            else:
                logger.critical(f"Unknown error, cannot fetch selected object.")

    @classmethod
    def multipleSelectFromTable(cls,
                                data: Sequence[SelectableInterface],
                                object_type: Optional[Type] = None,
                                default: Optional[str] = None) -> Sequence[SelectableInterface]:
        """Return a list of object (implementing SelectableInterface) selected by the user
        Raise IndexError of the data list is empty."""
        cls.__isInteractionAllowed()
        result = []
        pool = cast(List[SelectableInterface], data).copy()
        if object_type is None and len(pool) > 0:
            object_type = type(pool[0])
        if object_type is ExegolContainer:
            object_subject = "container"
        elif object_type is ExegolImage:
            object_subject = "image"
        else:
            object_subject = "object"
        while True:
            selected = cast(SelectableInterface, cls.selectFromTable(pool, object_type, default))
            result.append(selected)
            pool.remove(selected)
            if len(pool) == 0:
                return result
            elif not Confirm(f"Do you want to select another {object_subject}?", default=False):
                return result

    @classmethod
    def selectFromList(cls,
                       data: Union[Dict[str, str], List[str]],
                       subject: str = "an option",
                       title: str = "Options",
                       default: Optional[str] = None) -> Union[str, Tuple[str, str]]:
        """if data is list(str): Return a string selected by the user
        if data is dict: list keys and return a tuple of the selected key corresponding value
        Raise IndexError of the data list is empty."""
        cls.__isInteractionAllowed()
        if len(data) == 0:
            logger.warning("No options were found")
            raise IndexError
        if type(data) is dict:
            submit_data = list(data.keys())
        else:
            submit_data = cast(List[str], data)
        cls.printTable(submit_data, title=title)
        if default is None:
            default = submit_data[0]
        choice = Prompt.ask(f"[bold blue][?][/bold blue] Select {subject}", default=default, choices=submit_data,
                            show_choices=False)
        if type(data) is dict:
            return choice, data[choice]
        else:
            return choice

    @classmethod
    def printContainerRecap(cls, container: ExegolContainerTemplate) -> None:
        """
        Build and print a rich table with every configuration of the container
        :param container: Exegol container to print the table of
        :return:
        """
        # Load the image status if it is not already set.
        container.image.autoLoad()

        recap = cls.__buildContainerRecapTable(container)

        logger.empty_line()
        console.print(recap)
        logger.empty_line()

    @staticmethod
    def __buildContainerRecapTable(container: ExegolContainerTemplate) -> Table:
        """
        Build a rich table to recap in detail the configuration of a specified ExegolContainerTemplate or ExegolContainer
        :param container: The container to fetch config from
        :return: A rich table fully built
        """
        # Fetch data
        devices = container.config.getTextDevices(logger.isEnabledFor(ExeLog.VERBOSE))
        envs = container.config.getTextEnvs(logger.isEnabledFor(ExeLog.VERBOSE))
        ports = container.config.getTextPorts()
        sysctls = container.config.getSysctls()
        capabilities = container.config.getCapabilities()
        volumes = container.config.getTextMounts(logger.isEnabledFor(ExeLog.VERBOSE))
        creation_date = container.config.getTextCreationDate()
        comment = container.config.getComment()
        passwd = container.config.getPasswd()

        # Color code
        privilege_color = "bright_magenta"
        path_color = "magenta"

        # Build table
        recap = Table(border_style="grey35", box=box.SQUARE, title_justify="left", show_header=True)
        recap.title = "[not italic]:white_medium_star: [/not italic][gold3][g]Container summary[/g][/gold3]"
        # Header
        recap.add_column(f"[bold blue]Name[/bold blue]{os.linesep}[bold blue]Image[/bold blue]", justify="right")
        container_status = container.getTextStatus()

        container_info_header = (f"{container.getDisplayName()} {'(' + container_status + ')' if container_status else ''}{os.linesep}"
                                 f"{container.image.getName()}")
        if "N/A" not in container.image.getImageVersion():
            container_info_header += f" - v.{container.image.getImageVersion()}"
        if "Unknown" not in container.image.getStatus():
            container_info_header += f" ({container.image.getStatus(include_version=False)})"
        if container.image.getArch().split('/')[0] != EnvInfo.arch or logger.isEnabledFor(ExeLog.VERBOSE):
            color = ConsoleFormat.getArchColor(container.image.getArch())
            container_info_header += f" [{color}]({container.image.getArch()})[/{color}]"
        recap.add_column(container_info_header)
        # Main features
        if comment:
            recap.add_row("[bold blue]Comment[/bold blue]", comment)
        if passwd:
            recap.add_row(f"[bold blue]Credentials[/bold blue]", f"[deep_sky_blue3]{container.config.getUsername()}[/deep_sky_blue3] : [deep_sky_blue3]{passwd}[/deep_sky_blue3]")
        recap.add_row("[bold blue]Remote Desktop[/bold blue]", container.config.getDesktopConfig())
        if creation_date:
            recap.add_row("[bold blue]Creation date[/bold blue]", creation_date)
        recap.add_row("[bold blue]Console GUI[/bold blue]", boolFormatter(container.config.isGUIEnable()) + container.config.getTextGuiSockets())
        recap.add_row("[bold blue]Network[/bold blue]", container.config.getTextNetworkMode())
        recap.add_row("[bold blue]Timezone[/bold blue]", boolFormatter(container.config.isTimezoneShared()))
        recap.add_row("[bold blue]Exegol resources[/bold blue]", boolFormatter(container.config.isExegolResourcesEnable()) +
                      f"{'[bright_black](/opt/resources)[/bright_black]' if container.config.isExegolResourcesEnable() else ''}")
        recap.add_row("[bold blue]My resources[/bold blue]", boolFormatter(container.config.isMyResourcesEnable()) +
                      f"{'[bright_black]({})[/bright_black]'.format(container.config.getMyResourcesPath()) if container.config.isMyResourcesEnable() else ''}")
        recap.add_row("[bold blue]Shell logging[/bold blue]", boolFormatter(container.config.isShellLoggingEnable()) +
                      f"{'[bright_black](/workspace/logs)[/bright_black]' if container.config.isShellLoggingEnable() else ''}")
        if "N/A" not in container.config.getVpnName():
            recap.add_row("[bold blue]VPN[/bold blue]", container.config.getVpnName())
        if container.config.getPrivileged() is True:
            recap.add_row("[bold blue]Privileged[/bold blue]", '[orange3]On :fire:[/orange3]')
        else:
            recap.add_row("[bold blue]Privileged[/bold blue]", "[green]Off :heavy_check_mark:[/green]")
        if len(capabilities) > 0:
            recap.add_row(f"[bold blue]Capabilities[/bold blue]",
                          f"[{privilege_color}]{', '.join(capabilities)}[/{privilege_color}]")
        if container.config.isWorkspaceCustom():
            recap.add_row("[bold blue]Workspace[/bold blue]",
                          f'[{path_color}]{container.config.getHostWorkspacePath()}[/{path_color}] [bright_black](/workspace)[/bright_black]')
        else:
            recap.add_row("[bold blue]Workspace[/bold blue]", '[bright_magenta]Dedicated[/bright_magenta] [bright_black](/workspace)[/bright_black]')
        if len(devices) > 0:
            recap.add_row("[bold blue]Devices[/bold blue]", devices.strip())
        if len(envs) > 0:
            recap.add_row("[bold blue]Envs[/bold blue]", envs.strip())
        if len(ports) > 0:
            recap.add_row("[bold blue]Ports[/bold blue]", ports.strip())
        if len(volumes) > 0:
            recap.add_row("[bold blue]Volumes[/bold blue]", volumes.strip())
        if len(sysctls) > 0:
            recap.add_row("[bold blue]Systctls[/bold blue]", os.linesep.join(
                [f"[{privilege_color}]{key}[/{privilege_color}] = {getColor(value)[0]}{value}{getColor(value)[1]}" for
                 key, value in sysctls.items()]))
        return recap

    @classmethod
    def __isInteractionAllowed(cls) -> None:
        # if not ParametersManager().interactive_mode:  # TODO improve non-interactive mode
        #    logger.critical(f'A required information is missing. Exiting.')
        pass



from argparse import Namespace
from pathlib import Path
from typing import Tuple

from exegol.config.ConstantConfig import ConstantConfig
from exegol.config.DataCache import DataCache
from exegol.config.UserConfig import UserConfig
from exegol.manager.UpdateManager import UpdateManager
from exegol.utils.DockerUtils import DockerUtils


def ContainerCompleter(prefix: str, parsed_args: Namespace, **kwargs) -> Tuple[str, ...]:
    """Function to dynamically load a container list for CLI autocompletion purpose"""
    data = [c.name for c in DockerUtils().listContainers()]
    for obj in data:
        # filter data if needed
        if prefix and not obj.lower().startswith(prefix.lower()):
            data.remove(obj)
    return tuple(data)


def ImageCompleter(prefix: str, parsed_args: Namespace, **kwargs) -> Tuple[str, ...]:
    """Function to dynamically load an image list for CLI autocompletion purpose"""
    # Skip image completer when container hasn't been selected first (because parameters are all optional, parameters order is not working)
    if parsed_args is not None and str(parsed_args.action) == "start" and parsed_args.containertag is None:
        return ()
    try:
        if parsed_args is not None and str(parsed_args.action) == "install":
            data = [img_cache.name for img_cache in DataCache().get_images_data().data if img_cache.source == "remote"]
        else:
            data = [img_cache.name for img_cache in DataCache().get_images_data().data]
    except Exception as e:
        data = []
    if len(data) == 0:
        # Fallback with default data if the cache is not initialized yet
        data = ["full", "nightly", "ad", "web", "light", "osint"]
    for obj in data:
        # filter data if needed
        if prefix and not obj.lower().startswith(prefix.lower()):
            data.remove(obj)
    return tuple(data)


def HybridContainerImageCompleter(prefix: str, parsed_args: Namespace, **kwargs) -> Tuple[str, ...]:
    """Hybrid completer for auto-complet. The selector on exec action is hybrid between image and container depending on the mode (tmp or not).
    This completer will supply the adequate data."""
    # "exec" parameter is filled first before the selector argument
    # If "selector" is null but the selector parameter is set in the first exec slot, no longer need to supply completer options
    if parsed_args.selector is None and parsed_args.exec is not None and len(parsed_args.exec) > 0:
        return ()
    # In "tmp" mode, the user must choose an image, otherwise it's a container
    if parsed_args.tmp:
        return ImageCompleter(prefix, parsed_args, **kwargs)
    else:
        return ContainerCompleter(prefix, parsed_args, **kwargs)


def BuildProfileCompleter(prefix: str, parsed_args: Namespace, **kwargs) -> Tuple[str, ...]:
    """Completer function for build profile parameter. The completer must be trigger only when an image name have already been chosen."""
    # The build profile completer must be trigger only when an image name have been set by user
    if parsed_args is not None and parsed_args.imagetag is None:
        return ()

    # Handle custom build path
    if parsed_args is not None and parsed_args.build_path is not None:
        custom_build_path = Path(parsed_args.build_path).expanduser().absolute()
        # Check if we have a directory or a file to select the project directory
        if not custom_build_path.is_dir():
            custom_build_path = custom_build_path.parent
        build_path = custom_build_path
    else:
        # Default build path
        build_path = Path(UserConfig().exegol_images_path)

    # Check if directory path exist
    if not build_path.is_dir():
        return tuple()

    # Find profile list
    data = list(UpdateManager.listBuildProfiles(profiles_path=build_path).keys())
    for obj in data:
        if prefix and not obj.lower().startswith(prefix.lower()):
            data.remove(obj)
    return tuple(data)


def DesktopConfigCompleter(prefix: str, **kwargs) -> Tuple[str, ...]:
    options = list(UserConfig.desktop_available_proto)
    for obj in options:
        if prefix and not obj.lower().startswith(prefix.lower()):
            options.remove(obj)
    # TODO add interface enum
    return tuple(options)


def VoidCompleter(**kwargs) -> Tuple:
    """No option to auto-complet"""
    return ()



import io
import os
import shutil
import subprocess
import sys
import time
from pathlib import Path
from typing import Optional

from exegol.config.EnvInfo import EnvInfo
from exegol.console.ExegolPrompt import Confirm
from exegol.exceptions.ExegolExceptions import CancelOperation
from exegol.utils.ExeLog import logger, console


class GuiUtils:
    """This utility class allows determining if the current system supports the GUI
    from the information of the system (through X11 sharing)."""

    __distro_name = ""
    default_x11_path = "/tmp/.X11-unix"

    @classmethod
    def isX11GuiAvailable(cls) -> bool:
        """
        Check if the host OS can support GUI application with X11 sharing
        :return: bool
        """
        # GUI (X11 sharing) was not supported on Windows before WSLg
        if EnvInfo.isWindowsHost():
            return cls.__windowsGuiChecks()
        elif EnvInfo.isMacHost():
            return cls.__macGuiChecks()
        # Linux default is True
        return True

    @classmethod
    def isWaylandGuiAvailable(cls) -> bool:
        """
        Check if the host OS can support GUI application with WAYLAND sharing
        :return: bool
        """
        if EnvInfo.isWindowsHost():
            return False
        elif EnvInfo.isMacHost():
            return False
        return EnvInfo.isWaylandAvailable()

    @classmethod
    def getX11SocketPath(cls) -> Optional[str]:
        """
        Get the host path of the X11 socket
        :return:
        """
        if cls.__distro_name:
            # Distro name can only be set if the current host OS is Windows
            return f"\\\\wsl.localhost\\{cls.__distro_name}\\mnt\\wslg\\.X11-unix"
        elif EnvInfo.isWindowsHost():
            if EnvInfo.current_platform == "WSL":
                # Mount point from a WSL shell context
                return f"/mnt/wslg/.X11-unix"
            else:
                # From a Windows context, a WSL distro should have been supply during GUI (X11 sharing) checks
                logger.debug(f"No WSL distro have been previously found: '{cls.__distro_name}'")
                raise CancelOperation("Exegol tried to create a container with X11 sharing on a Windows host "
                                      "without having performed the availability tests before.")
        elif EnvInfo.isMacHost():
            # Docker desktop don't support UNIX socket through volume, we are using XQuartz over the network until then
            return None
        # Other distributions (Linux / Mac) have the default socket path
        return cls.default_x11_path

    @classmethod
    def getWaylandSocketPath(cls) -> Optional[Path]:
        """
        Get the host path of the Wayland socket
        :return:
        """
        wayland_dir = os.getenv("XDG_RUNTIME_DIR")
        wayland_socket = os.getenv("WAYLAND_DISPLAY")
        if wayland_dir is None or wayland_socket is None:
            return None
        return Path(wayland_dir, wayland_socket)

    @classmethod
    def getDisplayEnv(cls) -> str:
        """
        Get the current DISPLAY environment to access X11 socket
        :return:
        """
        if EnvInfo.isMacHost():
            # xquartz Mac mode
            return "host.docker.internal:0"

        # Add ENV check is case of user don't have it, which will mess up GUI (X11 sharing) if fallback does not work
        # @see https://github.com/ThePorgs/Exegol/issues/148
        if not EnvInfo.is_windows_shell:
            if os.getenv("DISPLAY") is None:
                logger.warning("The DISPLAY environment variable is not set on your host. This can prevent GUI apps to start through X11 sharing")

        # DISPLAY var is fetch from the current user environment. If it doesn't exist, using ':0'.
        return os.getenv('DISPLAY', ":0")

    @classmethod
    def getWaylandEnv(cls) -> str:
        """
        Get the current WAYLAND_DISPLAY environment to access wayland socket
        :return:
        """
        return os.getenv('WAYLAND_DISPLAY', 'wayland-0')

    # # # # # # Mac specific methods # # # # # #

    @classmethod
    def __macGuiChecks(cls) -> bool:
        """
        Procedure to check if the Mac host supports GUI (X11 sharing) with docker through XQuartz
        :return: bool
        """
        if not cls.__isXQuartzInstalled():
            logger.warning("Display sharing is [orange3]not supported[/orange3] on your mac without XQuartz installed. "
                           "You need to manually install [turquoise2]XQuartz[/turquoise2] and check the configuration 'Allow connections from network clients'.")
            return False
        logger.debug("XQuartz detected.")
        if not cls.__xquartzAllowNetworkClients():
            # Notify user to change configuration
            logger.error("XQuartz does not allow network connections. "
                         "You need to manually change the configuration to 'Allow connections from network clients'")
            # Add sys.platform check to exclude windows env (fix for mypy static code analysis)
            if sys.platform != "win32" and os.getuid() == 0:
                logger.warning("You are running exegol as [red]root[/red]! The root user cannot check in the user context whether XQuartz is properly configured or not.")
            return False

        # Check if XQuartz is started, check is dir exist and if there is at least one socket
        if not cls.__isXQuartzRunning():
            if not cls.__startXQuartz():
                logger.warning("Unable to start XQuartz service.")
                return False

        # The /tmp config is not necessary until you can use the unix socket with docker-desktop volume
        # Check if Docker Desktop is configured with /tmp in Docker Desktop > Preferences > Resources > File Sharing
        #if EnvInfo.isDockerDesktop() and not cls.__checkDockerDesktopResourcesConfig():
        #    logger.warning("Display sharing not possible, Docker Desktop configuration is incorrect. Please add /tmp in "
        #                   "[magenta]Docker Desktop > Preferences > Resources > File Sharing[/magenta]")
        #    return False
        return True

    @staticmethod
    def __checkDockerDesktopResourcesConfig() -> bool:
        """
            Check if Docker Desktop for macOS is configured correctly, allowing
            /tmp to be bind mounted into Docker containers, which is needed to
             mount /tmp/.X11-unix for display sharing.
             Return True if the configuration is correct and /tmp is part of the whitelisted resources
        """
        # Function not used for now because the X11 socket cannot be used for now with Docker Desktop
        docker_config = EnvInfo.getDockerDesktopResources()
        logger.debug(f"Docker Desktop configuration filesharingDirectories: {docker_config}")
        return '/tmp' in docker_config

    @staticmethod
    def __isXQuartzInstalled() -> bool:
        return 'xquartz' in os.getenv('DISPLAY', "").lower()

    @staticmethod
    def __xquartzAllowNetworkClients() -> bool:
        # defaults read org.xquartz.X11.plist nolisten_tcp
        conf_check = subprocess.run(["defaults", "read", "org.xquartz.X11.plist", "nolisten_tcp"],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
        logger.debug(f"XQuartz nolisten_tcp config: '{conf_check.stdout.strip().decode('utf-8')}'")
        return conf_check.stdout.strip() == b'0'

    @classmethod
    def __isXQuartzRunning(cls) -> bool:
        """
        Check if xquartz service is up by testing sockets
        """
        socket_path = Path(cls.default_x11_path)
        socket_x11_found = False
        if socket_path.is_dir():
            for file in socket_path.glob("*"):
                if file.is_socket():
                    socket_x11_found = True
                    break
        return socket_x11_found

    @staticmethod
    def __startXQuartz() -> bool:
        xhost_path = shutil.which("xhost")
        if not xhost_path:
            logger.error("xhost command not found, check your XQuartz installation")
            return False
        # Starting xquartz
        logger.debug("Starting XQuartz using xhost command")
        with console.status(f"Starting [green]XQuartz[/green]...", spinner_style="blue"):
            run_xhost = subprocess.run([xhost_path], shell=True,
                                       stdout=subprocess.DEVNULL,
                                       stderr=subprocess.DEVNULL)
        return run_xhost.returncode == 0

    # # # # # # Windows specific methods # # # # # #

    @classmethod
    def __windowsGuiChecks(cls) -> bool:
        """
        Procedure to check if the Windows host supports GUI (X11 sharing) with docker through WSLg
        :return: bool
        """
        logger.debug("Testing WSLg availability")
        # WSL + WSLg must be available on the Windows host for the GUI to work through X11 sharing
        if not cls.__wsl_available():
            if sys.platform != "win32" and os.getuid() == 0:
                logger.critical("You are running exegol as [red]root[/red]! The root user cannot be used to run Exegol on a Windows environment.")
            logger.error("WSL is [orange3]not available[/orange3] on your system. X11 sharing is not supported.")
            return False
        logger.debug("WSL is [green]available[/green] on the local system")
        # Only WSL2 support WSLg
        if EnvInfo.getDockerEngine() != EnvInfo.DockerEngine.WLS2:
            logger.debug(f"Docker current engine: {EnvInfo.getDockerEngine().value}")
            logger.error("Docker must be run with [orange3]WSL2[/orange3] engine in order to support X11 sharing (i.e. GUI apps).")
            return False
        logger.debug("Docker is using [green]WSL2[/green]")
        # X11 socket can only be shared from a WSL (to find WSLg mount point)
        if EnvInfo.current_platform != "WSL":
            logger.debug("Exegol is running from a Windows context (e.g. Powershell), a WSL instance must be found to share the WSLg X11 socket")
            cls.__distro_name = cls.__find_wsl_distro()
            logger.debug(f"Set WSL Distro as: '{cls.__distro_name}'")
            # If no WSL is found, propose to continue without GUI (X11 sharing)
            if not cls.__distro_name and not Confirm(
                    "Do you want to continue [orange3]without[/orange3] X11 sharing (i.e. GUI support)?", default=True):
                raise KeyboardInterrupt
        else:
            logger.debug("Using current WSL context for X11 socket sharing")
        if cls.__wslg_installed():
            logger.debug("WSLg seems to be installed.")
            return True
        elif cls.__wslg_eligible():
            logger.info("[green]WSLg[/green] is available on your system but [orange3]not installed[/orange3].")
            logger.info("Make sure, your Windows is [green]up-to-date[/green] and [green]WSLg[/green] is installed on "
                        "your host by running 'wsl --update' as [orange3]admin[/orange3].")
            return False
        logger.debug("WSLg is [orange3]not available[/orange3]")
        logger.warning("Display sharing is [orange3]not supported[/orange3] on your version of Windows. "
                       "You need to upgrade to [turquoise2]Windows 10+[/turquoise2].")
        return False

    @staticmethod
    def __wsl_test(path, name: Optional[str] = "docker-desktop") -> bool:
        """
        Check presence of a file in the WSL docker-desktop image.
        the targeted WSL image can be changed with 'name' parameter.
        If name is None, the default WSL image will be use.
        """
        if EnvInfo.isWindowsHost():
            wsl = shutil.which("wsl.exe")
            if not wsl:
                logger.warning("wsl.exe seems to be unavailable on your system.")
                return False
            if name is None:
                logger.debug(f"Running: wsl.exe test -f {path}")
                ret = subprocess.run(["wsl.exe", "test", "-f", path])
            else:
                logger.debug(f"Running: wsl.exe test -d {name} -f {path}")
                ret = subprocess.run(["wsl.exe", "-d", name, "test", "-f", path])
            return ret.returncode == 0
        logger.debug("Trying to run a WSL test without Windows?")
        return False

    @classmethod
    def __check_wsl_docker_integration(cls, distrib_name) -> bool:
        """
        Check the presence of the docker binary in the supplied WSL distribution.
        This test allows checking if docker integration is enabled.
        """
        return cls.__wsl_test("/usr/bin/docker", distrib_name)

    @classmethod
    def __wsl_available(cls) -> bool:
        """
        heuristic to detect if Windows Subsystem for Linux is available.

        Uses presence of /etc/os-release in the WSL image to say Linux is there.
        This is a de facto file standard across Linux distros.

        Tests the existence of WSL by searching in the default WSL first.
        However, if the default wsl is 'docker-desktop-data', the result will be false, so you have to test with docker-desktop.
        """
        if EnvInfo.isWindowsHost():
            wsl = shutil.which("wsl.exe")
            if not wsl:
                logger.debug("wsl.exe not found on the local system.")
                return False
            logger.debug("running: wsl.exe --status")
            ret = subprocess.Popen(["wsl.exe", "--status"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            ret.wait()
            if ret.returncode == 0:
                return True
            else:
                logger.debug(f"wsl.exe --status return code {ret.returncode}")
                logger.debug(str(ret.stdout))
                logger.debug(str(ret.stderr))
        logger.debug("WSL status command failed.. Trying a fallback check method.")
        return cls.__wsl_test("/etc/os-release", name=None) or cls.__wsl_test("/etc/os-release")

    @classmethod
    def __wslg_installed(cls) -> bool:
        """
        Check if WSLg is installed and deploy inside a WSL image by testing if the file wslg/versions.txt exist.
        :return: bool
        """
        if EnvInfo.current_platform == "WSL":
            if (Path("/mnt/host/wslg/versions.txt").is_file() or
                    Path("/mnt/wslg/versions.txt").is_file()):
                return True
            logger.debug("Unable to find WSLg locally.. Check /mnt/wslg/ or /mnt/host/wslg/")
        else:
            if (cls.__wsl_test("/mnt/host/wslg/versions.txt", name=cls.__distro_name) or
                    cls.__wsl_test("/mnt/wslg/versions.txt", name=cls.__distro_name)):
                return True
            logger.debug(f"Unable to find WSLg.. Check /mnt/wslg/ or /mnt/host/wslg/ on {cls.__distro_name}")
        logger.debug("WSLg check failed.. Trying a fallback check method.")
        return cls.__wsl_test("/mnt/host/wslg/versions.txt") or cls.__wsl_test("/mnt/wslg/versions.txt", name=None)

    @staticmethod
    def __wslg_eligible() -> bool:
        """
        Check if the current Windows version support WSLg
        :return:
        """
        if EnvInfo.current_platform == "WSL":
            # WSL is only available on Windows 10 & 11 so WSLg can be installed.
            return True
        try:
            os_version_raw, _, build_number_raw = EnvInfo.getWindowsRelease().split('.')[:3]
            os_version = int(os_version_raw)
        except ValueError:
            logger.debug(f"Impossible to find the version of windows: '{EnvInfo.getWindowsRelease()}'")
            logger.error("Exegol can't know if your [orange3]version of Windows[/orange3] can support dockerized GUIs (X11 sharing).")
            return False
        # Available for Windows 10 & 11
        if os_version >= 10:
            return True
        logger.debug(f"Current version of Windows doesn't support WSLg: {os_version_raw}.?.{build_number_raw}")
        return False

    @classmethod
    def __find_wsl_distro(cls) -> str:
        distro_name = ""
        # these distros cannot be used to load WSLg socket
        blacklisted_distro = ["docker-desktop", "docker-desktop-data"]
        logger.debug("Running: C:\\Windows\\system32\\wsl.exe -l")
        ret = subprocess.Popen(["C:\\Windows\\system32\\wsl.exe", "-l"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # Wait for WSL process to end
        ret.wait()
        if ret.returncode == 0:
            skip_header = True
            # parse distribs
            logger.debug("Found WSL distribution:")
            assert ret.stdout is not None
            for line in io.TextIOWrapper(ret.stdout, encoding="utf-16le"):
                # Skip WSL text header
                if skip_header:
                    skip_header = False
                    continue
                # Remove newline
                line = line.strip()
                # Skip if line is empty
                if not line:
                    continue
                # Remove default text message
                name = line.split()[0]
                logger.debug(f"- {name}")
                # Skip blacklisted WSL
                if name not in blacklisted_distro:
                    eligible = True
                    # Test if the current WSL has docker integration activated
                    while not cls.__check_wsl_docker_integration(name):
                        eligible = False
                        logger.warning(
                            f"The '{name}' WSL distribution can be used to [green]enable X11 sharing[/green] (i.e. GUI apps) on exegol but the docker integration is [orange3]not enabled[/orange3].")
                        if not Confirm(
                                f"Do you want to [red]manually[/red] enable docker integration for WSL '{name}'?",
                                default=True):
                            break
                        eligible = True
                    if eligible:
                        distro_name = name
                        break
            if distro_name:
                logger.verbose(f"Wsl '{distro_name}' distribution found, the WSLg service can be mounted in exegol.")
            else:
                logger.warning(
                    "No WSL distribution was found on your machine. At least one distribution must be available to allow Exegol to use WSLg.")
                if Confirm("Do you want Exegol to install one automatically (Ubuntu)?", default=True):
                    if cls.__create_default_wsl():
                        distro_name = "Ubuntu"
        else:
            assert ret.stderr is not None
            logger.error(
                f"Error while loading existing wsl distributions. {ret.stderr.read().decode('utf-16le')} (code: {ret.returncode})")
        return distro_name

    @classmethod
    def __create_default_wsl(cls) -> bool:
        logger.info("Creating Ubuntu WSL distribution. Please wait.")
        logger.debug("Running: C:\\Windows\\system32\\wsl.exe --install -d Ubuntu")
        ret = subprocess.Popen(["C:\\Windows\\system32\\wsl.exe", "--install", "-d", "Ubuntu"], stderr=subprocess.PIPE)
        ret.wait()
        logger.info("Please follow installation instructions on the new window.")
        if ret.returncode != 0:
            assert ret.stderr is not None
            logger.error(
                f"Error while install WSL Ubuntu: {ret.stderr.read().decode('utf-16le')} (code: {ret.returncode})")
            return False
        else:
            while not Confirm("Is the installation of Ubuntu [green]finished[/green]?", default=True):
                pass
            # Check if docker have default docker integration
            docker_settings = EnvInfo.getDockerDesktopSettings()
            if docker_settings is not None and docker_settings.get("EnableIntegrationWithDefaultWslDistro", docker_settings.get("enableIntegrationWithDefaultWslDistro", False)):
                logger.verbose("Set WSL Ubuntu as default to automatically enable docker integration")
                logger.debug("Running: C:\\Windows\\system32\\wsl.exe -s Ubuntu")
                # Set new WSL distribution as default to start it and enable docker integration
                ret = subprocess.Popen(["C:\\Windows\\system32\\wsl.exe", "-s", "Ubuntu"], stderr=subprocess.PIPE)
                ret.wait()
                # Wait for the docker integration (10 try, 1 sec apart)
                with console.status("Waiting for the activation of the docker integration", spinner_style="blue"):
                    for _ in range(10):
                        if cls.__check_wsl_docker_integration("Ubuntu"):
                            break
                        time.sleep(1)
            while not cls.__check_wsl_docker_integration("Ubuntu"):
                logger.error("The newly created WSL could not get the docker integration automatically. "
                             "It has to be activated [red]manually[/red]")
                logger.info("Enable WSL Docker integration for the newly created WSL in: [magenta]Docker Desktop > Settings > Resources > WSL Integration[/magenta]")
                if not Confirm("Has the WSL Ubuntu docker integration been [red]manually[/red] activated?",
                               default=True):
                    return False
            logger.success("WSL 'Ubuntu' successfully created with docker integration")
            return True



import logging
import os
from typing import Any, cast

from rich.console import Console
from rich.logging import RichHandler


# Customized logging class
class ExeLog(logging.Logger):
    """Project's Logger custom class"""
    # New logging level
    SUCCESS: int = 25
    VERBOSE: int = 15
    ADVANCED: int = 13

    @staticmethod
    def setVerbosity(verbose: int, quiet: bool = False):
        """Set logging level accordingly to the verbose count or with quiet enable."""
        if quiet:
            logger.setLevel(logging.CRITICAL)
        elif verbose == 1:
            logger.setLevel(ExeLog.VERBOSE)
        elif verbose == 2:
            logger.setLevel(ExeLog.ADVANCED)
        elif verbose >= 3:
            logger.setLevel(logging.DEBUG)
        else:
            # Default INFO
            logger.setLevel(logging.INFO)

    def debug(self, msg: Any, *args: Any, **kwargs: Any) -> None:
        """Change default debug text format with rich color support"""
        super(ExeLog, self).debug("{}[D]{} {}".format("[bold yellow3]", "[/bold yellow3]", msg), *args, **kwargs)

    def advanced(self, msg: Any, *args: Any, **kwargs: Any) -> None:
        """Add advanced logging method with text format / rich color support"""
        if self.isEnabledFor(ExeLog.ADVANCED):
            self._log(ExeLog.ADVANCED,
                      "{}[A]{} {}".format("[bold yellow3]", "[/bold yellow3]", msg), args, **kwargs)

    def verbose(self, msg: Any, *args: Any, **kwargs: Any) -> None:
        """Add verbose logging method with text format / rich color support"""
        if self.isEnabledFor(ExeLog.VERBOSE):
            self._log(ExeLog.VERBOSE,
                      "{}[V]{} {}".format("[bold blue]", "[/bold blue]", msg), args, **kwargs)

    def raw(self, msg: Any, level=VERBOSE, markup=False, highlight=False, emoji=False, rich_parsing=False) -> None:
        """Add raw text logging, used for stream printing."""
        if rich_parsing:
            markup = True
            highlight = True
            emoji = True
        if self.isEnabledFor(level):
            if type(msg) is bytes:
                msg = msg.decode('utf-8', errors="ignore")
            # Raw message are print directly to the console bypassing logging system and auto formatting
            console.print(msg, end='', markup=markup, highlight=highlight, emoji=emoji)

    def info(self, msg: Any, *args: Any, **kwargs: Any) -> None:
        """Change default info text format with rich color support"""
        super(ExeLog, self).info("{}[*]{} {}".format("[bold blue]", "[/bold blue]", msg), *args, **kwargs)

    def warning(self, msg: Any, *args: Any, **kwargs: Any) -> None:
        """Change default warning text format with rich color support"""
        super(ExeLog, self).warning("{}[!]{} {}".format("[bold orange3]", "[/bold orange3]", msg), *args, **kwargs)

    def error(self, msg: Any, *args: Any, **kwargs: Any) -> None:
        """Change default error text format with rich color support"""
        super(ExeLog, self).error("{}[-]{} {}".format("[bold red]", "[/bold red]", msg), *args, **kwargs)

    def exception(self, msg: Any, *args: Any, **kwargs: Any) -> None:
        """Change default exception text format with rich color support"""
        super(ExeLog, self).exception("{}[x]{} {}".format("[bold red]", "[/bold red]", msg), *args, **kwargs)

    def critical(self, msg: Any, *args: Any, **kwargs: Any) -> None:
        """Change default critical text format with rich color support
        Add auto exit."""
        super(ExeLog, self).critical("{}[!]{} {}".format("[bold red]", "[/bold red]", msg), *args, **kwargs)
        exit(1)

    def success(self, msg: Any, *args: Any, **kwargs: Any) -> None:
        """Add success logging method with text format / rich color support"""
        if self.isEnabledFor(ExeLog.SUCCESS):
            self._log(ExeLog.SUCCESS,
                      "{}[+]{} {}".format("[bold green]", "[/bold green]", msg), args, **kwargs)

    def empty_line(self, log_level: int = logging.INFO) -> None:
        """Print an empty line."""
        self.raw(os.linesep, level=log_level)


# Global rich console object
console: Console = Console()

# Main logging default config
# Set default Logger class as ExeLog
logging.setLoggerClass(ExeLog)

# Add new level to the logging config
logging.addLevelName(ExeLog.VERBOSE, "VERBOSE")
logging.addLevelName(ExeLog.SUCCESS, "SUCCESS")
logging.addLevelName(ExeLog.ADVANCED, "ADVANCED")
# Logging setup using RichHandler and minimalist text format
logging.basicConfig(
    format="%(message)s",
    handlers=[RichHandler(rich_tracebacks=True,
                          show_time=False,
                          markup=True,
                          show_level=False,
                          show_path=False,
                          console=console)]
)

# Global logger object
logger: ExeLog = cast(ExeLog, logging.getLogger("main"))
# Default log level
logger.setLevel(logging.INFO)



import os
import shutil
import subprocess
import tempfile
from datetime import datetime
from typing import Optional, Dict, Sequence, Tuple, Union

from docker.errors import NotFound, ImageNotFound, APIError
from docker.models.containers import Container

from exegol.config.EnvInfo import EnvInfo
from exegol.console.ExegolPrompt import Confirm
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.model.ContainerConfig import ContainerConfig
from exegol.model.ExegolContainerTemplate import ExegolContainerTemplate
from exegol.model.ExegolImage import ExegolImage
from exegol.model.SelectableInterface import SelectableInterface
from exegol.utils.ContainerLogStream import ContainerLogStream
from exegol.utils.ExeLog import logger, console
from exegol.utils.GuiUtils import GuiUtils
from exegol.utils.imgsync.ImageScriptSync import ImageScriptSync


class ExegolContainer(ExegolContainerTemplate, SelectableInterface):
    """Class of an exegol container already create in docker"""

    def __init__(self, docker_container: Container, model: Optional[ExegolContainerTemplate] = None):
        logger.debug(f"Loading container: {docker_container.name}")
        self.__container: Container = docker_container
        self.__id: str = docker_container.id
        self.__xhost_applied = False
        if model is None:
            image_name = ""
            try:
                # Try to find the attached docker image
                docker_image = docker_container.image
            except ImageNotFound:
                # If it is not found, the user has probably forcibly deleted it manually
                logger.warning(f"Some images were forcibly removed by docker when they were used by existing containers!")
                logger.error(f"The '{docker_container.name}' containers might not work properly anymore and should also be deleted and recreated with a new image.")
                docker_image = None
                image_name = "[red bold]BROKEN[/red bold]"
            # Create Exegol container from an existing docker container
            super().__init__(docker_container.name,
                             config=ContainerConfig(docker_container),
                             image=ExegolImage(name=image_name, docker_image=docker_image),
                             hostname=docker_container.attrs.get('Config', {}).get('Hostname'),
                             new_container=False)
            self.image.syncContainerData(docker_container)
            # At this stage, the container image object has an unknown status because no synchronization with a registry has been done.
            # This could be done afterwards (with container.image.autoLoad()) if necessary because it takes time.
            self.__new_container = False
        else:
            # Create Exegol container from a newly created docker container with its object template.
            super().__init__(docker_container.name,
                             config=ContainerConfig(docker_container),
                             # Rebuild config from docker object to update workspace path
                             image=model.image,
                             hostname=model.config.hostname,
                             new_container=False)
            self.__new_container = True
        self.image.syncStatus()

    def __str__(self) -> str:
        """Default object text formatter, debug only"""
        return f"{self.getRawStatus()} - {super().__str__()}"

    def __getState(self) -> Dict:
        """Technical getter of the container status dict"""
        self.__container.reload()
        return self.__container.attrs.get("State", {})

    def getRawStatus(self) -> str:
        """Raw text getter of the container status"""
        return self.__getState().get("Status", "unknown")

    def getTextStatus(self) -> str:
        """Formatted text getter of the container status"""
        status = self.getRawStatus().lower()
        if status == "unknown":
            return "Unknown"
        elif status == "exited":
            return "[red]Stopped[/red]"
        elif status == "running":
            return "[green]Running[/green]"
        return f"[orange3]{status}[/orange3]"

    def isNew(self) -> bool:
        """Check if the container has just been created or not"""
        return self.__new_container

    def isRunning(self) -> bool:
        """Check is the container is running. Return bool."""
        return self.getRawStatus() == "running"

    def getFullId(self) -> str:
        """Container's id getter"""
        return self.__id

    def getId(self) -> str:
        """Container's short id getter"""
        return self.__container.short_id

    def getKey(self) -> str:
        """Universal unique key getter (from SelectableInterface)"""
        return self.name

    def start(self) -> None:
        """Start the docker container"""
        if not self.isRunning():
            logger.info(f"Starting container {self.name}")
            self.__start_container()
            self.__postStartSetup()

    def __start_container(self) -> None:
        """
        This method start the container and display startup status update to the user.
        :return:
        """
        with console.status(f"Waiting to start {self.name}", spinner_style="blue") as progress:
            start_date = datetime.now()
            try:
                self.__container.start()
            except APIError as e:
                logger.debug(e)
                logger.critical(f"Docker raise a critical error when starting the container [green]{self.name}[/green], error message is: {e.explanation}")
            if not self.config.legacy_entrypoint:  # TODO improve startup compatibility check
                try:
                    # Try to find log / startup messages. Will time out after 2 seconds if the image don't support status update through container logs.
                    for line in ContainerLogStream(self.__container, start_date=start_date, timeout=2):
                        # Once the last log "READY" is received, the startup sequence is over and the execution can continue
                        if line == "READY":
                            break
                        elif line.startswith('[W]'):
                            line = line.replace('[W]', '')
                            logger.warning(line)
                        elif line.startswith('[E]'):
                            line = line.replace('[E]', '')
                            logger.error(line)
                        else:
                            logger.verbose(line)
                        progress.update(status=f"[blue][Startup][/blue] {line}")
                except KeyboardInterrupt:
                    # User can cancel startup logging with ctrl+C
                    logger.warning("User skip startup status updates. Spawning a shell now.")

    def stop(self, timeout: int = 10) -> None:
        """Stop the docker container"""
        if self.isRunning():
            logger.info(f"Stopping container {self.name}")
            with console.status(f"Waiting to stop ({timeout}s timeout)", spinner_style="blue"):
                self.__container.stop(timeout=timeout)

    def spawnShell(self) -> None:
        """Spawn a shell on the docker container"""
        self.__check_start_version()
        logger.info(f"Location of the exegol workspace on the host : {self.config.getHostWorkspacePath()}")
        for device in self.config.getDevices():
            logger.info(f"Shared host device: {device.split(':')[0]}")
        logger.success(f"Opening shell in Exegol '{self.name}'")
        # In case of multi-user environment, xhost must be set before opening each session to be sure
        self.__applyX11ACLs()
        # Using system command to attach the shell to the user terminal (stdin / stdout / stderr)
        envs = self.config.getShellEnvs()
        options = ""
        if len(envs) > 0:
            options += f" -e {' -e '.join(envs)}"
        cmd = f"docker exec{options} -ti {self.getFullId()} {self.config.getShellCommand()}"
        logger.debug(f"Opening shell with: {cmd}")
        os.system(cmd)
        # Docker SDK doesn't support (yet) stdin properly
        # result = self.__container.exec_run(ParametersManager().shell, stdout=True, stderr=True, stdin=True, tty=True,
        #                                    environment=self.config.getShellEnvs())
        # logger.debug(result)

    def exec(self, command: Union[str, Sequence[str]], as_daemon: bool = True, quiet: bool = False, is_tmp: bool = False) -> None:
        """Execute a command / process on the docker container.
        Set as_daemon to not follow the command stream and detach the execution
        Set quiet to disable logs message
        Set is_tmp if the container will automatically be removed after execution"""
        if not self.isRunning():
            self.start()
        if not quiet:
            logger.info("Executing command on Exegol")
            if logger.getEffectiveLevel() > logger.VERBOSE and not ParametersManager().daemon:
                logger.info("Hint: use verbose mode to see command output (-v).")
        exec_payload, str_cmd = ExegolContainer.formatShellCommand(command, quiet)
        stream = self.__container.exec_run(exec_payload, environment={"CMD": str_cmd, "DISABLE_AUTO_UPDATE": "true"}, detach=as_daemon, stream=not as_daemon)
        if as_daemon and not quiet:
            logger.success("Command successfully executed in background")
        else:
            try:
                # stream[0] : exit code
                # stream[1] : text stream
                for log in stream[1]:
                    if type(log) is bytes:
                        log = log.decode("utf-8")
                    logger.raw(log)
                if not quiet:
                    logger.success("End of the command")
            except KeyboardInterrupt:
                if not quiet and not is_tmp:
                    logger.info("Detaching process logging")
                    logger.warning("Exiting this command does [red]NOT[/red] stop the process in the container")

    @staticmethod
    def formatShellCommand(command: Union[str, Sequence[str]], quiet: bool = False, entrypoint_mode: bool = False) -> Tuple[str, str]:
        """Generic method to format a shell command and support zsh aliases.
        Set quiet to disable any logging here.
        Set entrypoint_mode to start the command with the entrypoint.sh config loader.
        - The first return argument is the payload to execute with every pre-routine for zsh.
        - The second return argument is the command itself in str format."""
        # Using base64 to escape special characters
        str_cmd = command if type(command) is str else ' '.join(command)
        # str_cmd = str_cmd.replace('"', '\\"')  # This fix shoudn' be necessary plus it can alter data like passwd
        if not quiet:
            logger.success(f"Command received: {str_cmd}")
        # ZSH pre-routine: Load zsh aliases and call eval to force aliases interpretation
        cmd = f'autoload -Uz compinit; compinit; source ~/.zshrc; eval "$CMD"'
        if not entrypoint_mode:
            # For direct execution, the full command must be supplied not just the zsh argument
            cmd = f"zsh -c '{cmd}'"
        return cmd, str_cmd

    def remove(self) -> None:
        """Stop and remove the docker container"""
        self.__removeVolume()
        self.stop(timeout=2)
        logger.info(f"Removing container {self.name}")
        try:
            self.__container.remove()
            logger.success(f"Container {self.name} successfully removed.")
        except NotFound:
            logger.error(
                f"The container {self.name} has already been removed (probably created as a temporary container).")

    def __removeVolume(self) -> None:
        """Remove private workspace volume directory if exist"""
        volume_path = self.config.getPrivateVolumePath()
        # TODO add backup
        if volume_path != '':
            if volume_path.startswith('/wsl/') or volume_path.startswith('\\wsl\\'):
                # Docker volume defines from WSL don't return the real path, they cannot be automatically removed
                # TODO review WSL workspace volume
                logger.warning("Warning: WSL workspace directory cannot be removed automatically.")
                return
            logger.verbose("Removing workspace volume")
            logger.debug(f"Removing volume {volume_path}")
            try:
                list_files = os.listdir(volume_path)
            except PermissionError:
                if Confirm(f"Insufficient permission to view workspace files {volume_path}, "
                           f"do you still want to delete them?", default=False):
                    # Set list_files as empty to skip user prompt again
                    list_files = []
                else:
                    return
            except FileNotFoundError:
                logger.debug("This workspace has already been removed.")
                return
            try:
                if len(list_files) > 0:
                    # Directory is not empty
                    if not Confirm(f"Workspace [magenta]{volume_path}[/magenta] is not empty, do you want to delete it?",
                                   default=False):
                        # User can choose not to delete the workspace on the host
                        return
                # Try to remove files from the host with user permission (work only without sub-directory)
                shutil.rmtree(volume_path)
            except PermissionError:
                logger.info(f"Deleting the workspace files from the [green]{self.name}[/green] container as root")
                # If the host can't remove the container's file and folders, the rm command is exec from the container itself as root
                self.exec("rm -rf /workspace", as_daemon=False, quiet=True)
                try:
                    shutil.rmtree(volume_path)
                except PermissionError:
                    logger.warning(f"I don't have the rights to remove [magenta]{volume_path}[/magenta] (do it yourself)")
                    return
            except Exception as err:
                logger.error(err)
                return
            logger.success("Private workspace volume removed successfully")

    def __postStartSetup(self) -> None:
        """
        Operation to be performed after starting a container
        :return:
        """
        self.__applyX11ACLs()

    def __check_start_version(self) -> None:
        """
        Check spawn.sh up-to-date status and update the script if needed
        :return:
        """
        # Up-to-date container have the script shared over a volume
        # But legacy container must be checked and the code must be pushed
        if not self.config.isWrapperStartShared():
            # If the spawn.sh if not shared, the version must be compared and the script updated
            current_start = ImageScriptSync.getCurrentStartVersion()
            # Try to parse the spawn version of the container. If an alpha or beta version is in use, the script will always be updated.
            spawn_parsing_cmd = ["/bin/bash", "-c", "egrep '^# Spawn Version:[0-9]+$' /.exegol/spawn.sh 2&>/dev/null || echo ':0' | cut -d ':' -f2"]
            container_version = self.__container.exec_run(spawn_parsing_cmd).output.decode("utf-8").strip()
            if current_start != container_version:
                logger.debug(f"Updating spawn.sh script from version {container_version} to version {current_start}")
                self.__container.put_archive("/", ImageScriptSync.getImageSyncTarData(include_spawn=True))

    def postCreateSetup(self, is_temporary: bool = False) -> None:
        """
        Operation to be performed after creating a container
        :return:
        """
        self.__applyX11ACLs()
        # if not a temporary container, apply custom config
        if not is_temporary:
            # Update entrypoint script in the container
            self.__container.put_archive("/", ImageScriptSync.getImageSyncTarData(include_entrypoint=True))
            if self.__container.status.lower() == "created":
                self.__start_container()
            try:
                self.__updatePasswd()
            except APIError as e:
                if "is not running" in e.explanation:
                    logger.critical("An unexpected error occurred. Exegol cannot start the container after its creation...")

    def __applyX11ACLs(self) -> None:
        """
        If X11 (GUI) is enabled, allow X11 access on host ACL (if not already allowed) for linux and mac.
        If the host is accessed by SSH, propagate xauth cookie authentication if applicable.
        On Windows host, WSLg X11 don't have xhost ACL. #TODO xauth remote x11 forwarding
        :return:
        """
        # TODO check if the xauth propagation should be performed on Windows, if so remove the "and not EnvInfo.isWindowsHost()"
        if self.config.isGUIEnable() and not self.__xhost_applied and not EnvInfo.isWindowsHost():
            self.__xhost_applied = True  # Can be applied only once per execution
            if shutil.which("xhost") is None:
                if EnvInfo.is_linux_shell:
                    debug_msg = "Try to install the package [green]xorg-xhost[/green] or maybe you don't have X11 on your host?"
                else:
                    debug_msg = "or you don't have one"
                logger.error(f"The [green]xhost[/green] command is not available on your [bold]host[/bold]. "
                             f"Exegol was unable to allow your container to access your graphical environment ({debug_msg}).")
                return

            logger.debug(f"DISPLAY variable: {GuiUtils.getDisplayEnv()}")
            # Extracts the left part of the display variable to determine if remote access is used
            display_host = GuiUtils.getDisplayEnv().split(':')[0]
            # Left part is empty, local access is used to start Exegol
            if display_host == '' or EnvInfo.isMacHost():
                logger.debug("Connecting to container from local GUI, no X11 forwarding to set up")
                # TODO verify that the display format is the same on macOS, otherwise might not set up xauth and xhost correctly
                if EnvInfo.isMacHost():
                    logger.debug(f"Adding xhost ACL to localhost")
                    # add xquartz inet ACL
                    with console.status(f"Starting XQuartz...", spinner_style="blue"):
                        os.system(f"xhost + localhost > /dev/null")
                elif not EnvInfo.isWindowsHost():
                    logger.debug(f"Adding xhost ACL to local:{self.config.getUsername()}")
                    # add linux local ACL
                    os.system(f"xhost +local:{self.config.getUsername()} > /dev/null")
                return

            if shutil.which("xauth") is None:
                if EnvInfo.is_linux_shell:
                    debug_msg = "Try to install the package [green]xorg-xauth[/green] to support X11 forwarding in your current environment?"
                else:
                    debug_msg = "or it might not be supported for now"
                logger.error(f"The [green]xauth[/green] command is not available on your [bold]host[/bold]. "
                             f"Exegol was unable to allow your container to access your graphical environment ({debug_msg}).")
                return

            # If the left part of the display variable is "localhost", x11 socket is exposed only on loopback and remote access is used
            # If the container is not in host mode, it won't be able to reach the loopback interface of the host
            if display_host == "localhost" and self.config.getNetworkMode() != "host":
                logger.warning("X11 forwarding won't work on a bridged container unless you specify \"X11UseLocalhost no\" in your host sshd_config")
                logger.warning("[red]Be aware[/red] changing \"X11UseLocalhost\" value can [red]expose your device[/red], correct firewalling is [red]required[/red]")
                # TODO Add documentation to restrict the exposure of the x11 socket to the docker subnet
                return

            # Extracting the xauth cookie corresponding to the current display to a temporary file and reading it from there (grep cannot be used because display names are not accurate enough)
            _, tmpXauthority = tempfile.mkstemp()
            logger.debug(f"Extracting xauth entries to {tmpXauthority}")
            os.system(f"xauth extract {tmpXauthority} $DISPLAY > /dev/null 2>&1")
            xauthEntry = subprocess.check_output(f"xauth -f {tmpXauthority} list 2>/dev/null", shell=True).decode()
            logger.debug(f"xauthEntry to propagate: {xauthEntry}")

            # Replacing the hostname with localhost to support loopback exposed x11 socket and container in host mode (loopback is the same)
            if display_host == "localhost":
                logger.debug("X11UseLocalhost directive is set to \"yes\" or unspecified, X11 connections can be received only on loopback");
                # Modifing the entry to convert <hostname>/unix:<display_number> to localhost:<display_number>
                xauthEntry = f"localhost:{xauthEntry.split(':')[1]}"
            else:
                # TODO latter implement a check to see if the x11 socket is correctly firewalled and warn the user if it is not
                logger.debug("X11UseLocalhost directive is set to \"no\", X11 connections can be received from anywere");

            # Check if the host has a xauth entry corresponding to the current display.
            if xauthEntry:
                logger.debug(f"Adding xauth cookie to container: {xauthEntry}")
                self.exec(f"xauth add {xauthEntry}", as_daemon=False, quiet=True)
                logger.debug(f"Removing {tmpXauthority}")
                os.remove(tmpXauthority)
            else:
                logger.warning(f"No xauth cookie corresponding to the current display was found.")

    def __updatePasswd(self) -> None:
        """
        If configured, update the password of the user inside the container.
        :return:
        """
        if self.config.getPasswd() is not None:
            logger.debug(f"Updating the {self.config.getUsername()} password inside the container")
            self.exec(f"echo '{self.config.getUsername()}:{self.config.getPasswd()}' | chpasswd", quiet=True)



import json
import os
import platform
from enum import Enum
from pathlib import Path
from typing import Optional, List, Dict

from exegol.config.ConstantConfig import ConstantConfig
from exegol.utils.ExeLog import logger


class EnvInfo:
    """Class to identify the environment in which exegol runs to adapt
    the configurations, processes and messages for the user"""

    class HostOs(Enum):
        """Dictionary class for static OS Name"""
        WINDOWS = "Windows"
        LINUX = "Linux"
        MAC = "Mac"

    class DisplayServer(Enum):
        """Dictionary class for static Display Server"""
        WAYLAND = "Wayland"
        X11 = "X11"

    class DockerEngine(Enum):
        """Dictionary class for static Docker engine name"""
        WLS2 = "WSL2"
        HYPERV = "Hyper-V"
        DOCKER_DESKTOP = "Docker desktop"
        ORBSTACK = "Orbstack"
        LINUX = "Kernel"

    """Contain information about the environment (host, OS, platform, etc)"""
    # Shell env
    current_platform: str = "WSL" if "microsoft" in platform.release() else platform.system()  # Can be 'Windows', 'Linux' or 'WSL'
    is_linux_shell: bool = current_platform in ["WSL", "Linux"]
    is_windows_shell: bool = current_platform == "Windows"
    is_mac_shell: bool = not is_windows_shell and not is_linux_shell  # If not Linux nor Windows, its (probably) a mac
    __is_docker_desktop: bool = False
    __windows_release: Optional[str] = None
    # Host OS
    __docker_host_os: Optional[HostOs] = None
    __docker_engine: Optional[DockerEngine] = None
    # Docker desktop cache config
    __docker_desktop_resource_config: Optional[dict] = None
    # Architecture
    raw_arch = platform.machine().lower()
    arch = raw_arch
    if arch == "x86_64" or arch == "x86-64" or arch == "amd64":
        arch = "amd64"
    elif arch == "aarch64" or "armv8" in arch:
        arch = "arm64"
    elif "arm" in arch:
        if platform.architecture()[0] == '64bit':
            arch = "arm64"
        else:
            logger.error(f"Host architecture seems to be 32-bit ARM ({arch}), which is not supported yet. "
                         f"If possible, please install a 64-bit operating system (Exegol supports ARM64).")
        """
        if "v5" in arch:
            arch = "arm/v5"
        elif "v6" in arch:
            arch = "arm/v6"
        elif "v7" in arch:
            arch = "arm/v7"
        elif "v8" in arch:
            arch = "arm64"
        """
    else:
        logger.warning(f"Unknown / unsupported architecture: {arch}. Using 'AMD64' as default.")
        # Fallback to default AMD64 arch
        arch = "amd64"

    @classmethod
    def initData(cls, docker_info) -> None:
        """Initialize information from Docker daemon data"""
        # Fetch data from Docker daemon
        docker_os = docker_info.get("OperatingSystem", "unknown").lower()
        docker_kernel = docker_info.get("KernelVersion", "unknown").lower()
        # Deduct a Windows Host from data
        cls.__is_docker_desktop = docker_os == "docker desktop"
        is_host_windows = cls.__is_docker_desktop and "microsoft" in docker_kernel
        is_orbstack = (docker_os == "orbstack" or "(containerized)" in docker_os) and "orbstack" in docker_kernel
        if is_host_windows:
            # Check docker engine with Windows host
            if "wsl2" in docker_kernel:
                cls.__docker_engine = cls.DockerEngine.WLS2
            else:
                cls.__docker_engine = cls.DockerEngine.HYPERV
            cls.__docker_host_os = cls.HostOs.WINDOWS
        elif cls.__is_docker_desktop:
            # If docker desktop is detected but not a Windows engine/kernel, it's (probably) a mac
            cls.__docker_engine = cls.DockerEngine.DOCKER_DESKTOP
            cls.__docker_host_os = cls.HostOs.MAC if cls.is_mac_shell else cls.HostOs.LINUX
        elif is_orbstack:
            # Orbstack is only available on Mac
            cls.__docker_engine = cls.DockerEngine.ORBSTACK
            cls.__docker_host_os = cls.HostOs.MAC
        else:
            # Every other case it's a linux distro and docker is powered from the kernel
            cls.__docker_engine = cls.DockerEngine.LINUX
            cls.__docker_host_os = cls.HostOs.LINUX

        if cls.__docker_engine == cls.DockerEngine.DOCKER_DESKTOP and cls.__docker_host_os == cls.HostOs.LINUX:
            logger.warning(f"Using Docker Desktop on Linux is not officially supported !")

    @classmethod
    def getHostOs(cls) -> HostOs:
        """Return Host OS
        Can be 'Windows', 'Mac' or 'Linux'"""
        # initData must be called from DockerUtils on client initialisation
        assert cls.__docker_host_os is not None
        return cls.__docker_host_os

    @classmethod
    def getDisplayServer(cls) -> DisplayServer:
        """Returns the display server
        Can be 'X11' or 'Wayland'"""
        session_type = os.getenv("XDG_SESSION_TYPE", "x11")
        if session_type == "wayland":
            return cls.DisplayServer.WAYLAND
        elif session_type == "x11":
            return cls.DisplayServer.X11
        else:
            # Should return an error
            logger.warning(f"Unknown session type {session_type}. Using X11 as fallback.")
            return cls.DisplayServer.X11

    @classmethod
    def getWindowsRelease(cls) -> str:
        # Cache check
        if cls.__windows_release is None:
            if cls.is_windows_shell:
                # From a Windows shell, python supply an approximate (close enough) version of windows
                cls.__windows_release = platform.win32_ver()[1]
            else:
                cls.__windows_release = "Unknown"
        return cls.__windows_release

    @classmethod
    def isWindowsHost(cls) -> bool:
        """Return true if Windows is detected on the host"""
        return cls.getHostOs() == cls.HostOs.WINDOWS

    @classmethod
    def isMacHost(cls) -> bool:
        """Return true if macOS is detected on the host"""
        return cls.getHostOs() == cls.HostOs.MAC

    @classmethod
    def isLinuxHost(cls) -> bool:
        """Return true if Linux is detected on the host"""
        return cls.getHostOs() == cls.HostOs.LINUX

    @classmethod
    def isWaylandAvailable(cls) -> bool:
        """Return true if wayland is detected on the host"""
        return cls.getDisplayServer() == cls.DisplayServer.WAYLAND or bool(os.getenv("WAYLAND_DISPLAY"))

    @classmethod
    def isDockerDesktop(cls) -> bool:
        """Return true if docker desktop is used on the host"""
        return cls.__is_docker_desktop

    @classmethod
    def isOrbstack(cls) -> bool:
        """Return true if docker desktop is used on the host"""
        return cls.__docker_engine == cls.DockerEngine.ORBSTACK

    @classmethod
    def getDockerEngine(cls) -> DockerEngine:
        """Return Docker engine type.
        Can be any of EnvInfo.DockerEngine"""
        # initData must be called from DockerUtils on client initialisation
        assert cls.__docker_engine is not None
        return cls.__docker_engine

    @classmethod
    def getShellType(cls) -> str:
        """Return the type of shell exegol is executed from"""
        if cls.is_linux_shell:
            return cls.HostOs.LINUX.value
        elif cls.is_windows_shell:
            return cls.HostOs.WINDOWS.value
        elif cls.is_mac_shell:
            return cls.HostOs.MAC.value
        else:
            return "Unknown"

    @classmethod
    def getDockerDesktopSettings(cls) -> Dict:
        """Applicable only for docker desktop on macos"""
        if cls.isDockerDesktop():
            if cls.__docker_desktop_resource_config is None:
                dir_path = None
                file_path = None
                if cls.is_mac_shell:
                    # Mac PATH
                    dir_path = ConstantConfig.docker_desktop_mac_config_path
                elif cls.is_windows_shell:
                    # Windows PATH
                    dir_path = ConstantConfig.docker_desktop_windows_config_path
                else:
                    # Windows PATH from WSL shell
                    # Find docker desktop config
                    config_file = list(Path("/mnt/c/Users").glob(f"*/{ConstantConfig.docker_desktop_windows_config_short_path}/settings-store.json"))
                    if len(config_file) == 0:
                        # Testing with legacy file name
                        config_file = list(Path("/mnt/c/Users").glob(f"*/{ConstantConfig.docker_desktop_windows_config_short_path}/settings.json"))
                        if len(config_file) == 0:
                            logger.warning(f"No docker desktop settings file found.")
                            return {}
                    file_path = config_file[0]
                if file_path is None:
                    assert dir_path is not None
                    # Try to find settings file with new filename or fallback to legacy filename for Docker Desktop older than 4.34
                    file_path = (dir_path / "settings-store.json") if (dir_path / "settings-store.json").is_file() else (dir_path / "settings.json")
                logger.debug(f"Loading Docker Desktop config from {file_path}")
                try:
                    with open(file_path, 'r') as docker_desktop_config:
                        cls.__docker_desktop_resource_config = json.load(docker_desktop_config)
                except FileNotFoundError:
                    logger.warning(f"Docker Desktop configuration file not found: '{file_path}'")
                    return {}
            if cls.__docker_desktop_resource_config is None:
                logger.warning(f"Docker Desktop configuration couldn't be loaded.'")
            else:
                return cls.__docker_desktop_resource_config
        return {}

    @classmethod
    def getDockerDesktopResources(cls) -> List[str]:
        settings = cls.getDockerDesktopSettings()
        # Handle legacy settings key
        docker_desktop_resources = settings.get('FilesharingDirectories', settings.get('filesharingDirectories', []))
        logger.debug(f"Docker Desktop resources whitelist: {docker_desktop_resources}")
        return docker_desktop_resources

    @classmethod
    def isHostNetworkAvailable(cls) -> bool:
        if cls.isLinuxHost():
            return True
        elif cls.isOrbstack():
            return True
        elif cls.isDockerDesktop():
            settings = cls.getDockerDesktopSettings()
            # Handle legacy settings key
            res = settings.get('HostNetworkingEnabled', settings.get('hostNetworkingEnabled', False))
            return res if res is not None else False
        logger.warning("Unknown or not supported environment for host network mode.")
        return False



import os
from typing import Optional

from rich.prompt import Prompt

from exegol.config.EnvInfo import EnvInfo
from exegol.model.ContainerConfig import ContainerConfig
from exegol.model.ExegolImage import ExegolImage


class ExegolContainerTemplate:
    """Exegol template class used to create a new container"""

    def __init__(self, name: Optional[str], config: ContainerConfig, image: ExegolImage, hostname: Optional[str] = None, new_container: bool = True):
        if name is None:
            name = Prompt.ask("[bold blue][?][/bold blue] Enter the name of your new exegol container", default="default")
        assert name is not None
        if (EnvInfo.isWindowsHost() or EnvInfo.isMacHost()) and not name.startswith("exegol-"):
            # Force container as lowercase because the filesystem of windows / mac are case-insensitive => https://github.com/ThePorgs/Exegol/issues/167
            name = name.lower()
        self.container_name: str = name if name.startswith("exegol-") else f'exegol-{name}'
        self.name: str = name.replace('exegol-', '')
        self.image: ExegolImage = image
        self.config: ContainerConfig = config
        if hostname:
            self.config.hostname = hostname
            if new_container:
                self.config.addEnv(ContainerConfig.ExegolEnv.exegol_name.value, self.container_name)
        else:
            self.config.hostname = self.container_name

    def __str__(self) -> str:
        """Default object text formatter, debug only"""
        return f"{self.name} - {self.image.getName()}{os.linesep}{self.config}"

    def prepare(self) -> None:
        """Prepare the model before creating the docker container"""
        self.config.prepareShare(self.name)

    def rollback(self) -> None:
        """Rollback change in case of container creation fail."""
        self.config.rollback_preparation(self.name)

    def getDisplayName(self) -> str:
        """Getter of the container's name for TUI purpose"""
        if self.container_name != self.config.hostname:
            return f"{self.name} [bright_black]({self.config.hostname})[/bright_black]"
        return self.name

    def getTextStatus(self) -> str:
        return ""



from argparse import Namespace
from typing import List, Any

from exegol.console.cli.actions.Command import Command
from exegol.utils.ExeLog import logger
from exegol.utils.MetaSingleton import MetaSingleton
from exegol.utils.argParse import Parser


class ParametersManager(metaclass=MetaSingleton):
    """This class is a singleton allowing to access from anywhere to any parameter
    filled by the user from the CLI arguments"""

    def __init__(self) -> None:
        # List every action available on the project (from the root Class)
        actions: List[Command] = [cls() for cls in Command.__subclasses__()]
        # Load & execute argparse
        parser: Parser = Parser(actions)
        parsing_results = parser.run_parser()
        # The user arguments resulting from the parsing will be stored in parameters
        self.parameters: Command = self.__loadResults(parser, parsing_results)

    @staticmethod
    def __loadResults(parser: Parser, parsing_results: Namespace) -> Command:
        """The result of argparse is sent to the action object to replace the parser with the parsed values"""
        try:
            action: Command = parsing_results.action
            action.populate(parsing_results)
            return action
        except AttributeError:
            # Catch missing "action" parameter en CLI
            parser.print_help()
            exit(0)

    def getCurrentAction(self) -> Command:
        """Return the object corresponding to the action selected by the user"""
        return self.parameters

    def __getattr__(self, item: str) -> Any:
        """The getattr function is overloaded to transparently pass the parameter search
        in the child object of Command stored in the 'parameters' attribute"""
        try:
            # The priority is to first return the attributes of the current object
            # Using the object generic method to avoid infinite loop to itself
            return object.__getattribute__(self, item)
        except AttributeError:
            # If parameters is called before initialisation (from the next statement), this can create an infinite loop
            if item == "parameters":
                return None
        try:
            # If item was not found in self, the search is initiated among the parameters
            return getattr(self.parameters, item)
        except AttributeError:
            # The logger may not work if the call is made before its initialization
            logger.debug(f"Attribute not found in parameters: {item}")
            return None

    def __setattr__(self, key, value) -> None:
        """Allow to dynamically change some parameter during runtime"""
        # Only some specific parameters are whitelisted for runtime update
        if key in ["offline_mode"]:
            setattr(self.parameters, key, value)
        else:
            super().__setattr__(key, value)



from pathlib import Path
from typing import List

from exegol.config.ConstantConfig import ConstantConfig
from exegol.console.ConsoleFormat import boolFormatter
from exegol.utils.DataFileUtils import DataFileUtils
from exegol.utils.MetaSingleton import MetaSingleton


class UserConfig(DataFileUtils, metaclass=MetaSingleton):
    """This class allows loading user defined configurations"""

    # Static choices
    start_shell_options = {'zsh', 'bash', 'tmux'}
    shell_logging_method_options = {'script', 'asciinema'}
    desktop_available_proto = {'http', 'vnc'}

    def __init__(self) -> None:
        # Defaults User config
        self.private_volume_path: Path = ConstantConfig.exegol_config_path / "workspaces"
        self.my_resources_path: Path = ConstantConfig.exegol_config_path / "my-resources"
        self.exegol_resources_path: Path = self.__default_resource_location('exegol-resources')
        self.exegol_images_path: Path = self.__default_resource_location('exegol-images')
        self.auto_check_updates: bool = True
        self.auto_remove_images: bool = True
        self.auto_update_workspace_fs: bool = False
        self.default_start_shell: str = "zsh"
        self.enable_exegol_resources: bool = True
        self.shell_logging_method: str = "asciinema"
        self.shell_logging_compress: bool = True
        self.desktop_default_enable: bool = False
        self.desktop_default_localhost: bool = True
        self.desktop_default_proto: str = "http"

        super().__init__("config.yml", "yml")

    def _build_file_content(self) -> str:
        config = f"""# Exegol configuration
# Full documentation: https://exegol.readthedocs.io/en/latest/exegol-wrapper/advanced-uses.html#id1

# Volume path can be changed at any time but existing containers will not be affected by the update
volumes:
    # The my-resources volume is a storage space dedicated to the user to customize his environment and tools. This volume can be shared across all exegol containers.
    # Attention! The permissions of this folder (and subfolders) will be updated to share read/write rights between the host (user) and the container (root). Do not modify this path to a folder on which the permissions (chmod) should not be modified.
    my_resources_path: {self.my_resources_path}
    
    # Exegol resources are data and static tools downloaded in addition to docker images. These tools are complementary and are accessible directly from the host.
    exegol_resources_path: {self.exegol_resources_path}
    
    # Exegol images are the source of the exegol environments. These sources are needed when locally building an exegol image.
    exegol_images_path: {self.exegol_images_path}
    
    # When containers do not have an explicitly declared workspace, a dedicated folder will be created at this location to share the workspace with the host but also to save the data after deleting the container
    private_workspace_path: {self.private_volume_path}

config:
    # Enables automatic check for wrapper updates
    auto_check_update: {self.auto_check_updates}
    
    # Automatically remove outdated image when they are no longer used
    auto_remove_image: {self.auto_remove_images}
    
    # Automatically modifies the permissions of folders and sub-folders in your workspace by default to enable file sharing between the container with your host user.
    auto_update_workspace_fs: {self.auto_update_workspace_fs}
    
    # Default shell command to start
    default_start_shell: {self.default_start_shell}
    
    # Enable Exegol resources
    enable_exegol_resources: {self.enable_exegol_resources}
    
    # Change the configuration of the shell logging functionality
    shell_logging:
        #Choice of the method used to record the sessions (script or asciinema)
        logging_method: {self.shell_logging_method}
        
        # Enable automatic compression of log files (with gzip)
        enable_log_compression: {self.shell_logging_compress}
        
    # Configure your Exegol Desktop
    desktop:
        # Enables or not the desktop mode by default
        # If this attribute is set to True, then using the CLI --desktop option will be inverted and will DISABLE the feature
        enabled_by_default: {self.desktop_default_enable}
        
        # Default desktop protocol,can be "http", or "vnc" (additional protocols to come in the future, check online documentation for updates).
        default_protocol: {self.desktop_default_proto}
        
        # Desktop service is exposed on localhost by default. If set to true, services will be exposed on localhost (127.0.0.1) otherwise it will be exposed on 0.0.0.0. This setting can be overwritten with --desktop-config
        localhost_by_default: {self.desktop_default_localhost}

"""
        return config

    @staticmethod
    def __default_resource_location(folder_name: str) -> Path:
        local_src = ConstantConfig.src_root_path_obj / folder_name
        if local_src.is_dir():
            # If exegol is clone from github, exegol submodule is accessible from root src
            return local_src
        else:
            # Default path for pip installation
            return ConstantConfig.exegol_config_path / folder_name

    def _process_data(self) -> None:
        # Volume section
        volumes_data = self._raw_data.get("volumes", {})
        # Catch existing but empty section
        if volumes_data is None:
            volumes_data = {}
        self.my_resources_path = self._load_config_path(volumes_data, 'my_resources_path', self.my_resources_path)
        self.private_volume_path = self._load_config_path(volumes_data, 'private_workspace_path', self.private_volume_path)
        self.exegol_resources_path = self._load_config_path(volumes_data, 'exegol_resources_path', self.exegol_resources_path)
        self.exegol_images_path = self._load_config_path(volumes_data, 'exegol_images_path', self.exegol_images_path)

        # Config section
        config_data = self._raw_data.get("config", {})
        # Catch existing but empty section
        if config_data is None:
            config_data = {}
        self.auto_check_updates = self._load_config_bool(config_data, 'auto_check_update', self.auto_check_updates)
        self.auto_remove_images = self._load_config_bool(config_data, 'auto_remove_image', self.auto_remove_images)
        self.auto_update_workspace_fs = self._load_config_bool(config_data, 'auto_update_workspace_fs', self.auto_update_workspace_fs)
        self.default_start_shell = self._load_config_str(config_data, 'default_start_shell', self.default_start_shell, choices=self.start_shell_options)
        self.enable_exegol_resources = self._load_config_bool(config_data, 'enable_exegol_resources', self.enable_exegol_resources)

        # Shell_logging section
        shell_logging_data = config_data.get("shell_logging", {})
        self.shell_logging_method = self._load_config_str(shell_logging_data, 'logging_method', self.shell_logging_method, choices=self.shell_logging_method_options)
        self.shell_logging_compress = self._load_config_bool(shell_logging_data, 'enable_log_compression', self.shell_logging_compress)

        # Desktop section
        desktop_data = config_data.get("desktop", {})
        self.desktop_default_enable = self._load_config_bool(desktop_data, 'enabled_by_default', self.desktop_default_enable)
        self.desktop_default_proto = self._load_config_str(desktop_data, 'default_protocol', self.desktop_default_proto, choices=self.desktop_available_proto)
        self.desktop_default_localhost = self._load_config_bool(desktop_data, 'localhost_by_default', self.desktop_default_localhost)

    def get_configs(self) -> List[str]:
        """User configs getter each options"""
        configs = [
            f"User config file: [magenta]{self._file_path}[/magenta]",
            f"Private workspace: [magenta]{self.private_volume_path}[/magenta]",
            "Exegol resources: " + (f"[magenta]{self.exegol_resources_path}[/magenta]"
                                    if self.enable_exegol_resources else
                                    boolFormatter(self.enable_exegol_resources)),
            f"Exegol images: [magenta]{self.exegol_images_path}[/magenta]",
            f"My resources: [magenta]{self.my_resources_path}[/magenta]",
            f"Auto-check updates: {boolFormatter(self.auto_check_updates)}",
            f"Auto-remove images: {boolFormatter(self.auto_remove_images)}",
            f"Auto-update fs: {boolFormatter(self.auto_update_workspace_fs)}",
            f"Default start shell: [blue]{self.default_start_shell}[/blue]",
            f"Shell logging method: [blue]{self.shell_logging_method}[/blue]",
            f"Shell logging compression: {boolFormatter(self.shell_logging_compress)}",
            f"Desktop enabled by default: {boolFormatter(self.desktop_default_enable)}",
            f"Desktop default protocol: [blue]{self.desktop_default_proto}[/blue]",
            f"Desktop default host: [blue]{'localhost' if self.desktop_default_localhost else '0.0.0.0'}[/blue]",
        ]
        # TUI can't be called from here to avoid circular importation
        return configs
